"""

Simulate localization data.

This module provides functions to simulate localization data and return LocData objects.
Localizations are often distributed either by a spatial process of complete-spatial randomness or following a
Neyman-Scott process [1]_. For a Neyman-Scott process parent events (representing single emitters) yield a random number
of cluster_mu events (representing localizations due to repeated blinking). Related spatial point processes include
Matérn and Thomas processes.

Functions that are named as make_* provide point data arrays. Functions that are named as simulate_* provide
locdata.


Parts of this code is adapted from scikit-learn/sklearn/datasets/_samples_generator.py .
(BSD 3-Clause License, Copyright (c) 2007-2020 The scikit-learn developers.)

References
----------
.. [1] Neyman, J. & Scott, E. L.,
   A Theory of the Spatial Distribution of Galaxies.
   Astrophysical Journal 1952, vol. 116, p.144.

"""
import sys
import time
from itertools import chain

import numpy as np
import pandas as pd

from locan.data.locdata import LocData
from locan.data import metadata_pb2
from locan.data.region import Region, EmptyRegion, Interval, Rectangle, Ellipse, \
    AxisOrientedCuboid, AxisOrientedHypercuboid
from locan.data.region_utils import expand_region


__all__ = [
           'make_uniform', 'make_Poisson', 'make_cluster', 'make_NeymanScott', 'make_Matern', 'make_Thomas',
           'make_dstorm',
           'simulate_uniform', 'simulate_Poisson',
           'simulate_cluster', 'simulate_NeymanScott', 'simulate_Matern', 'simulate_Thomas',
           'simulate_dstorm',
           'simulate_tracks', 'resample', 'simulate_frame_numbers'
]


def make_uniform(n_samples, region=(0, 1), seed=None):
    """
    Provide points that are distributed by a uniform (complete spatial randomness) point process
    within the boundaries given by `region`.

    Parameters
    ----------
    n_samples : int
        The total number of localizations of the point process
    region : Region, array-like
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        Random number generation seed

    Returns
    -------
    numpy.ndarray of shape (n_samples, n_features)
        The generated samples.
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    if isinstance(region, EmptyRegion):
        samples = np.array([])
    elif isinstance(region, (Interval, Rectangle, AxisOrientedCuboid, AxisOrientedHypercuboid)):
        samples = rng.uniform(region.bounds[:region.dimension], region.bounds[region.dimension:],
                              size=(n_samples, region.dimension))
    elif isinstance(region, Ellipse) and region.width == region.height:
        radius = region.width / 2
        # angular and radial coordinates of Poisson points
        theta = rng.random(n_samples) * 2 * np.pi
        rho = radius * np.sqrt(rng.random(n_samples))
        # Convert from polar to Cartesian coordinates
        xx = rho * np.cos(theta)
        yy = rho * np.sin(theta)
        samples = np.array((xx, yy)).T
    else:
        sampling_ratio = region.region_measure / region.bounding_box.region_measure
        n_samples_updated = int(n_samples / sampling_ratio * 2)

        samples = []
        n_remaining = n_samples
        while n_remaining > 0:
            new_samples = rng.random(size=(n_samples_updated, region.dimension))
            new_samples = region.extent * new_samples + region.bounding_box.corner
            new_samples = new_samples[region.contains(new_samples)]
            samples.append(new_samples)
            n_remaining = n_remaining - len(new_samples)

        samples = np.concatenate(samples)
        samples = samples[0: n_samples]

    return samples


def simulate_uniform(n_samples, region=(0, 1), seed=None):
    """
    Provide points that are distributed by a uniform Poisson point process within the boundaries given by `region`.

    Parameters
    ----------
    n_samples : int
        The total number of localizations of the point process
    region : Region, array-like
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples = make_uniform(n_samples=n_samples, region=region, seed=seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_uniform.__name__, parameter=str(parameter))

    return locdata


def make_Poisson(intensity, region=(0, 1), seed=None):
    """
    Provide points that are distributed by a uniform Poisson point process within the boundaries given by `region`.

    Parameters
    ----------
    intensity : int, float
        The intensity (points per unit region measure) of the point process
    region : Region, array-like
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    numpy.ndarray of shape (n_samples, n_features)
        The generated samples.
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    n_samples = rng.poisson(lam=intensity*region.region_measure)

    if isinstance(region, EmptyRegion):
        samples = np.array([])
    elif n_samples == 0:
        samples = np.array([])
        for i in range(region.dimension):
            samples = samples[..., np.newaxis]
    elif isinstance(region, (Interval, Rectangle, AxisOrientedCuboid, AxisOrientedHypercuboid)):
        samples = rng.uniform(region.bounds[:region.dimension], region.bounds[region.dimension:],
                              size=(n_samples, region.dimension))
    elif isinstance(region, Ellipse) and region.width == region.height:
        radius = region.width / 2
        # angular and radial coordinates of Poisson points
        theta = rng.random(n_samples) * 2 * np.pi
        rho = radius * np.sqrt(rng.random(n_samples))
        # Convert from polar to Cartesian coordinates
        xx = rho * np.cos(theta)
        yy = rho * np.sin(theta)
        samples = np.array((xx, yy)).T
    else:
        sampling_ratio = region.region_measure / region.bounding_box.region_measure
        n_samples_updated = int(n_samples / sampling_ratio * 2)

        samples = []
        n_remaining = n_samples
        while n_remaining > 0:
            new_samples = rng.random(size=(n_samples_updated, region.dimension))
            new_samples = region.extent * new_samples + region.bounding_box.corner
            new_samples = new_samples[region.contains(new_samples)]
            samples.append(new_samples)
            n_remaining = n_remaining - len(new_samples)

        samples = np.concatenate(samples)
        samples = samples[0: n_samples]

    return samples


def simulate_Poisson(intensity, region=(0, 1), seed=None):
    """
    Provide points that are distributed by a uniform Poisson point process within the boundaries given by `region`.

    Parameters
    ----------
    intensity : int, float
        The intensity (points per unit region measure) of the point process
    region : Region, array-like
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples = make_Poisson(intensity=intensity, region=region, seed=seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_Poisson.__name__, parameter=str(parameter))

    return locdata


def make_cluster(centers=3, region=(0, 1.), expansion_distance=0, offspring=None,
                 clip=True, shuffle=True, seed=None):
    """
    Parent positions are taken from `centers`
    or are distributed according to a homogeneous Poisson process with exactly `centers`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by cluster_mu offspring points as passed or generated by a given function.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    centers : int, array-like
        The number of parents or coordinates for parent events, where each parent represents a cluster center.
    region : Region, array-like
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance : float
        The distance by which region is expanded on all boundaries.
    offspring : array-like, callable, None
        Points or function for point process to provide cluster.
        Callable must take single parent point as parameter and return an iterable.
        If array-like it must have the same length as parent events.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    tuple of numpy.ndarray of shape (n_samples, n_features) and region
       The generated samples, labels, parent_samples, region
    """
    rng = np.random.default_rng(seed)

    if isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = np.array([]), np.array([]), np.array([]), EmptyRegion()
        return samples, labels, parent_samples, region
    elif not isinstance(region, Region):
        region = Region.from_intervals(region)

    expanded_region = expand_region(region, expansion_distance)

    if isinstance(centers, (int, np.integer)):
        n_centers = centers
        parent_samples = make_uniform(n_samples=n_centers, region=expanded_region, seed=rng)
    else:  # if centers is array
        parent_samples = np.array(centers)
        centers_shape = np.shape(parent_samples)
        centers_dimension = 1 if len(centers_shape) == 1 else centers_shape[1]
        if region.dimension != centers_dimension:
            raise ValueError(f'Region dimensions must be the same as the dimensions for each center. '
                             f'Got region dimension: {region.dimension} and '
                             f'center dimensions: {centers_dimension} instead.')
        n_centers = centers_shape[0]

    # replace parents by offspring samples
    if offspring is None:
        samples = parent_samples
        labels = np.arange(0, len(parent_samples))

    elif callable(offspring):
        try:
            offspring_samples = offspring(parent_samples)
            labels = [[i] * len(os) for i, os in enumerate(offspring_samples)]
        except TypeError:
            offspring_samples = []
            labels = []
            for i, parent in enumerate(parent_samples):
                offspring_samples_ = offspring(parent)
                offspring_samples.append(offspring_samples_)
                labels.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels)))

    elif len(offspring) >= len(parent_samples):
        offspring_samples = []
        labels = []
        for i, (os, parent) in enumerate(zip(offspring[:n_centers], parent_samples)):
            if len(os) > 0:
                offspring_samples_ = np.asarray(os) + parent
                offspring_samples.append(offspring_samples_)
                labels.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels)))

    else:
        raise TypeError(f"offspring must be callable or array-like with length >= than n_centers {n_centers}.")

    if samples.ndim == 1:  # this is to convert 1-dimensional arrays into arrays with shape (n_samples, 1).
        samples.shape = (len(samples), 1)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if len(samples) == 0:  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_cluster(centers=3, region=(0, 1.), expansion_distance=0, offspring=None,
                     clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data.
    Parent positions are taken from `centers`
    or are distributed according to a homogeneous Poisson process with exactly `centers`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by offspring points as passed or generated by a given function.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    centers : int, array-like
        The number of parents or coordinates for parent events, where each parent represents a cluster center.
    region : Region, array-like
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance : float
        The distance by which region is expanded on all boundaries.
    offspring : array-like, callable, None
        Points or function for point process to provide cluster.
        Callable must take single parent point as parameter and return an iterable.
        If array-like it must have the same length as parent events.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_cluster(centers, region, expansion_distance, offspring,
                                     clip, shuffle, seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_cluster.__name__, parameter=str(parameter))

    return locdata


def make_NeymanScott(parent_intensity=100, region=(0, 1.), expansion_distance=0, offspring=None,
                     clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Neyman-Scott random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by offspring points as passed or generated by a given function.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance : float
        The distance by which region is expanded on all boundaries.
    offspring : array-like, callable, None
        Points or function for point process to provide offspring points.
        Callable must take single parent point as parameter.
        If array-like it must have enough elements to fit the randomly generated number of parent events.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    tuple of numpy.ndarray of shape (n_samples, n_features)
       The generated samples, labels, parent_samples
    """
    rng = np.random.default_rng(seed)

    if isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = np.array([]), np.array([]), np.array([]), EmptyRegion()
        return samples, labels, parent_samples, region
    elif not isinstance(region, Region):
        region = Region.from_intervals(region)

    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(intensity=parent_intensity, region=expanded_region, seed=rng)

    # replace parents by offspring samples
    if offspring is None:
        samples = parent_samples
        labels = np.arange(0, len(parent_samples))

    elif callable(offspring):
        try:
            offspring_samples = offspring(parent_samples)
            labels = [[i] * len(os) for i, os in enumerate(offspring_samples)]
        except TypeError:
            offspring_samples = []
            labels = []
            for i, parent in enumerate(parent_samples):
                offspring_samples_ = offspring(parent)
                offspring_samples.append(offspring_samples_)
                labels.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels)))

    elif len(offspring) >= len(parent_samples):
        offspring_samples = []
        labels = []
        if isinstance(offspring, np.ndarray):
            offspring_samples = np.asarray(offspring[:len(parent_samples)]) + parent_samples
            labels = [[i] * len(os) for i, os in enumerate(offspring_samples)]
        else:
            for i, (os, parent) in enumerate(zip(offspring[:len(parent_samples)], parent_samples)):
                if len(os) > 0:
                    offspring_samples_ = np.asarray(os) + parent
                    offspring_samples.append(offspring_samples_)
                    labels.append([i] * len(offspring_samples_))
        samples = np.array(list(chain(*offspring_samples)))
        labels = np.array(list(chain(*labels)))

    else:
        raise TypeError(f"offspring must be callable or array-like with "
                        f"length >= n_centers {len(parent_samples)}.")

    if samples.ndim == 1:  # this is to convert 1-dimensional arrays into arrays with shape (n_samples, 1).
        samples.shape = (len(samples), 1)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if len(samples) == 0:  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_NeymanScott(parent_intensity=100, region=(0, 1.), expansion_distance=0, offspring=None,
                     clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Neyman-Scott random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by the expansion_distance.
    Each parent position is then replaced by offspring points as passed or generated by a given function.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_distance : float
        The distance by which region is expanded on all boundaries.
    offspring : array-like, callable, None
        Points or function for point process to provide offspring points.
        Callable must take single parent point as parameter.
        If array-like it must have enough elements to fit the randomly generated number of parent events.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_NeymanScott(parent_intensity, region, expansion_distance, offspring,
                                     clip, shuffle, seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_NeymanScott.__name__, parameter=str(parameter))

    return locdata


def make_Matern(parent_intensity=1, region=(0, 1.), cluster_mu=1, radius=1.0,
                clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Matern cluster random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by the maximum radius.
    Each parent position is then replaced by spots of size `radius` with Poisson distributed points inside.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    cluster_mu : int, float
        The mean number of points of the Poisson point process for cluster(cluster_mu) events.
    radius : float or sequence of floats
        The radius for the spots. If tuple, the number of elements must be larger than the expected number of parents.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    tuple of numpy.ndarray of shape (n_samples, n_features)
       The generated samples, labels, parent_samples
    """
    rng = np.random.default_rng(seed)

    if isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = np.array([]), np.array([]), np.array([]), EmptyRegion()
        return samples, labels, parent_samples, region
    elif not isinstance(region, Region):
        region = Region.from_intervals(region)

    expansion_distance = np.max(radius)
    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(intensity=parent_intensity, region=expanded_region, seed=rng)

    # radius: if radius is given as list, it must be consistent with the n_parent_samples
    if hasattr(radius, "__len__"):
        if len(radius) < len(parent_samples):
            raise ValueError(f"Length of `radius` {len(radius)} is less than "
                             f"the generated n_parent_samples {len(parent_samples)}."
                             )
        else:
            radii = radius[:len(parent_samples)]
    else:  # if isinstance(radius, float):
        radii = np.full(len(parent_samples), radius)

    # replace parents by offspring samples
    samples = []
    labels = []
    for i, (parent, radius_) in enumerate(zip(parent_samples, radii)):
        if region.dimension == 1:
            offspring_region = Interval(-radius_, radius_)
            offspring_intensity = cluster_mu / offspring_region.region_measure
            offspring_samples = make_Poisson(intensity=offspring_intensity, region=offspring_region, seed=rng)
        elif region.dimension == 2:
            offspring_region = Ellipse((0, 0), 2*radius_, 2*radius_, 0)
            offspring_intensity = cluster_mu / offspring_region.region_measure
            offspring_samples = make_Poisson(intensity=offspring_intensity, region=offspring_region, seed=rng)
        elif region.dimension == 3:
            raise NotImplementedError
        else:
            raise ValueError("region dimension must be 1, 2, or 3.")
        if len(offspring_samples) != 0:
            offspring_samples = offspring_samples + parent
            samples.append(offspring_samples)
        labels += [i] * len(offspring_samples)

    samples = np.array(list(chain(*samples)))
    labels = np.array(labels)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if len(samples) == 0:  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_Matern(parent_intensity=1, region=(0, 1.), cluster_mu=1, radius=1.0,
                clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Matern cluster random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by the maximum radius.
    Each parent position is then replaced by spots of size `radius` with Poisson distributed points inside.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    cluster_mu : int, float
        The mean number of points of the Poisson point process for cluster(cluster_mu) events.
    radius : float or sequence of floats
        The radius for the spots. If tuple, the number of elements must be larger than the expected number of parents.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_Matern(parent_intensity, region, cluster_mu, radius,
                          clip, shuffle, seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_Matern.__name__, parameter=str(parameter))

    return locdata


def make_Thomas(parent_intensity=1, region=(0, 1.), expansion_factor=6,
                cluster_mu=1, cluster_std=1.0,
                clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Thomas random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is Poisson-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor : int, float
        Factor by which the cluster_std is multiplied to set the region expansion distance.
    cluster_mu : int, float, sequence of floats
        The mean number of points for normal-distributed offspring points.
    cluster_std : float, sequence of floats, sequence of sequence of floats
        The standard deviation for normal-distributed offspring points.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    tuple of numpy.ndarray of shape (n_samples, n_features)
       The generated samples, labels, parent_samples
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    if parent_intensity == 0 or \
            (np.size(cluster_mu) == 1 and cluster_mu == 0) or \
            isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = np.array([]), np.array([]), np.array([]), region
        if region.dimension and region.dimension > 0:
            samples = samples[:, np.newaxis]
        return samples, labels, parent_samples, region

    # expand region
    expansion_distance = expansion_factor * np.max(cluster_std)
    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(intensity=parent_intensity, region=expanded_region, seed=rng)
    n_cluster = len(parent_samples)

    # check cluster_std consistent with n_centers or n_features
    if len(np.shape(cluster_std)) == 0:
        cluster_std_ = np.full(shape=(n_cluster, region.dimension), fill_value=cluster_std)
    elif len(np.shape(cluster_std)) == 1:  # iterate over cluster_std for each feature
        if region.dimension == 1 or len(cluster_std) != region.dimension:
            raise TypeError(f"The shape of cluster_std {np.shape(cluster_std)} is incompatible "
                            f"with n_features {region.dimension}.")
        else:
            cluster_std_ = np.empty(shape=(n_cluster, region.dimension))
            for i, element in enumerate(cluster_std):
                cluster_std_[:, i] = np.full((n_cluster,), element)
    elif len(np.shape(cluster_std)) == 2:  # iterate over cluster_std for each center
        if np.shape(cluster_std) < (n_cluster, region.dimension):
            raise TypeError(f"The shape of cluster_std {np.shape(cluster_std)} is incompatible with "
                            f"n_cluster {n_cluster} or n_features {region.dimension}.")
        else:
            cluster_std_ = cluster_std
    else:
        raise TypeError(f"The shape of cluster_std {np.shape(cluster_std)} is incompatible.")

    # replace parents by normal-distributed offspring samples
    try:
        n_offspring_list = rng.poisson(lam=cluster_mu[:n_cluster], size=n_cluster)
    except ValueError as e:
        e.args += (f"Too few offspring events for n_cluster: {n_cluster}",)
        raise
    except (TypeError, IndexError):
        n_offspring_list = rng.poisson(lam=cluster_mu, size=n_cluster)
    samples = []
    labels = []
    for i, (parent, std, n_offspring) in enumerate(zip(parent_samples, cluster_std_, n_offspring_list)):
        offspring_samples = rng.normal(loc=parent, scale=std, size=(n_offspring, region.dimension))
        samples.append(offspring_samples)
        labels += [i] * len(offspring_samples)

    samples = np.concatenate(samples) if len(samples) != 0 else np.array([])
    labels = np.array(labels)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if len(samples) == 0:  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_Thomas(parent_intensity=1, region=(0, 1.), expansion_factor=6,
                cluster_mu=1, cluster_std=1.0,
                clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Thomas random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is Poisson-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor : int, float
        Factor by which the cluster_std is multiplied to set the region expansion distance.
    cluster_mu : int, float, sequence of floats
        The mean number of points for normal-distributed offspring points.
    cluster_std : float, sequence of floats, sequence of sequence of floats
        The standard deviation for normal-distributed offspring points.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_Thomas(parent_intensity, region, expansion_factor,
                          cluster_mu, cluster_std,
                          clip, shuffle, seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_Thomas.__name__, parameter=str(parameter))

    return locdata

def make_dstorm(parent_intensity=1, region=(0, 1.), expansion_factor=6,
                cluster_mu=1, min_points=0, cluster_std=1.0,
                clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Thomas-like random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is geometrically-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for all features.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor : int, float
        Factor by which the cluster_std is multiplied to set the region expansion distance.
    cluster_mu : int, float, sequence of floats
        The mean number of points for normal-distributed offspring points.
    min_points : int
        The minimum number of points per cluster.
    cluster_std : float, sequence of floats, sequence of sequence of floats
        The standard deviation for normal-distributed offspring points.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    tuple of numpy.ndarray of shape (n_samples, n_features)
       The generated samples, labels, parent_samples
    """
    rng = np.random.default_rng(seed)

    if not isinstance(region, Region):
        region = Region.from_intervals(region)

    if parent_intensity == 0 or \
            (np.size(cluster_mu) == 1 and cluster_mu == 0) or \
            isinstance(region, EmptyRegion):
        samples, labels, parent_samples, region = np.array([]), np.array([]), np.array([]), region
        if region.dimension and region.dimension > 0:
            samples = samples[:, np.newaxis]
        return samples, labels, parent_samples, region

    # expand region
    expansion_distance = expansion_factor * np.max(cluster_std)
    expanded_region = expand_region(region, expansion_distance)

    parent_samples = make_Poisson(intensity=parent_intensity, region=expanded_region, seed=rng)
    n_cluster = len(parent_samples)

    # check cluster_std consistent with n_centers or n_features
    if len(np.shape(cluster_std)) == 0:
        cluster_std_ = np.full(shape=(n_cluster, region.dimension), fill_value=cluster_std)
    elif len(np.shape(cluster_std)) == 1:  # iterate over cluster_std for each feature
        if region.dimension == 1 or len(cluster_std) != region.dimension:
            raise TypeError(f"The shape of cluster_std {np.shape(cluster_std)} is incompatible "
                            f"with n_features {region.dimension}.")
        else:
            cluster_std_ = np.empty(shape=(n_cluster, region.dimension))
            for i, element in enumerate(cluster_std):
                cluster_std_[:, i] = np.full((n_cluster,), element)
    elif len(np.shape(cluster_std)) == 2:  # iterate over cluster_std for each center
        if np.shape(cluster_std) < (n_cluster, region.dimension):
            raise TypeError(f"The shape of cluster_std {np.shape(cluster_std)} is incompatible with "
                            f"n_cluster {n_cluster} or n_features {region.dimension}.")
        else:
            cluster_std_ = cluster_std
    else:
        raise TypeError(f"The shape of cluster_std {np.shape(cluster_std)} is incompatible.")

    # replace parents by normal-distributed offspring samples
    try:
        p_values = [1 / (mu + 1 - min_points) for mu in cluster_mu[:n_cluster]]
        # p for a geometric distribution sampling points from min_points to inf is 1 / (mean + 1 - min_points)
        n_offspring_list = rng.geometric(p=p_values, size=n_cluster) - 1 + min_points
        # rng.geometric samples points from 1 to inf.
    except ValueError as e:
        e.args += (f"Too few offspring events for n_cluster: {n_cluster}",)
        raise
    except (TypeError, IndexError):
        n_offspring_list = rng.geometric(p=1 / (cluster_mu + 1 - min_points),
                                         size=n_cluster) - 1 + min_points

    samples = []
    labels = []
    for i, (parent, std, n_offspring) in enumerate(zip(parent_samples, cluster_std_, n_offspring_list)):
        offspring_samples = rng.normal(loc=parent, scale=std, size=(n_offspring, region.dimension))
        samples.append(offspring_samples)
        labels += [i] * len(offspring_samples)

    samples = np.concatenate(samples) if len(samples) != 0 else np.array([])
    labels = np.array(labels)

    if clip is True:
        if len(samples) != 0:
            inside_indices = region.contains(samples)
            samples = samples[inside_indices]
            labels = labels[inside_indices]
        region_ = region
    else:
        region_ = expanded_region

    if shuffle:
        shuffled_indices = rng.permutation(len(samples))
        samples = samples[shuffled_indices]
        labels = labels[shuffled_indices]

    if len(samples) == 0:  # this is to convert empty arrays into arrays with shape (n_samples, n_features).
        samples = np.array([])
        samples = samples[:, np.newaxis]

    return samples, labels, parent_samples, region_


def simulate_dstorm(parent_intensity=1, region=(0, 1.), expansion_factor=6,
                cluster_mu=1, min_points=0, cluster_std=1.0,
                clip=True, shuffle=True, seed=None):
    """
    Generate clustered point data following a Thomas-like random point process.
    Parent positions are distributed according to a homogeneous Poisson process with `parent_intensity`
    within the boundaries given by `region` expanded by an expansion distance that equals
    expansion_factor * max(cluster_std).
    Each parent position is then replaced by n offspring points
    where n is geometrically-distributed with mean number `cluster_mu`
    and point coordinates are normal-distributed around the parent point with standard deviation `cluster_std`.
    Offspring from parent events that are located outside the region are included.

    Parameters
    ----------
    parent_intensity : int, float
        The intensity (points per unit region measure) of the Poisson point process for parent events.
    region : Region, array-like
        The region (or support) for each feature.
        If array-like it must provide upper and lower bounds for each feature.
    expansion_factor : int, float
        Factor by which the cluster_std is multiplied to set the region expansion distance.
    cluster_mu : int, float, sequence of floats
        The mean number of points for normal-distributed offspring points.
    min_points : int
        The minimum number of points per cluster.
    cluster_std : float, sequence of floats, sequence of sequence of floats
        The standard deviation for normal-distributed offspring points.
    clip : bool
        If True the result will be clipped to 'region'. If False the extended region will be kept.
    shuffle : boolean
        Shuffle the samples.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        The generated samples.
    """
    parameter = locals()
    samples, labels, _, region = make_dstorm(parent_intensity, region, expansion_factor,
                          cluster_mu, min_points, cluster_std,
                          clip, shuffle, seed)
    region_ = region if isinstance(region, Region) else Region.from_intervals(region)
    locdata = LocData.from_coordinates(coordinates=samples)
    locdata.dimension = region_.dimension
    locdata.region = region_
    locdata.dataframe = locdata.dataframe.assign(cluster_label=labels)

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=make_dstorm.__name__, parameter=str(parameter))

    return locdata


def _random_walk(n_walks=1, n_steps=10, dimensions=2, diffusion_constant=1, time_step=10, seed=None):
    """
    Random walk simulation

    Parameters
    ----------
    n_walks: int
        Number of walks
    n_steps : int
        Number of time steps (i.e. frames)
    dimensions : int
        spatial dimensions to simulate
    diffusion_constant : int or float
        Diffusion constant in units length per seconds^2
    time_step : float
        Time per frame (or simulation step) in seconds.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    tuple of arrays
        (times, positions), where shape(times) is 1 and shape of positions is (n_walks, n_steps, dimensions)
    """
    rng = np.random.default_rng(seed)
    # equally spaced time steps
    times = np.arange(n_steps) * time_step

    # random step sizes according to the diffusion constant
    random_numbers = rng.integers(0, 2, size=(n_walks, n_steps, dimensions))  # np.random.randint(0, 2, size=(n_walks, n_steps, dimensions))
    step_size = np.sqrt(2 * dimensions * diffusion_constant * time_step)
    steps = np.where(random_numbers == 0, -step_size, +step_size)

    # walker positions
    positions = np.cumsum(steps, axis=1)

    return times, positions


def simulate_tracks(n_walks=1, n_steps=10, ranges=((0, 10000), (0, 10000)), diffusion_constant=1,
                    time_step=10, seed=None):
    """
    Provide a dataset of localizations representing random walks with starting points being spatially-distributed
    on a rectangular shape or cubic volume by complete spatial randomness.

    Parameters
    ----------
    n_walks: int
        Number of walks
    n_steps : int
        Number of time steps (i.e. frames)
    ranges : tuple of tuples of two ints
        the range for valid x[, y, z]-coordinates
    diffusion_constant : int or float
        Diffusion constant with unit length per seconds^2
    time_step : float
        Time per frame (or simulation step) in seconds.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    LocData
        A new LocData instance with localization data.
    """
    parameter = locals()

    start_positions = np.array([np.random.uniform(*_range, size=n_walks) for _range in ranges]).T

    times, positions = _random_walk(n_walks=n_walks, n_steps=n_steps, dimensions=len(ranges),
                                    diffusion_constant=diffusion_constant, time_step=time_step, seed=seed)

    new_positions = np.concatenate([
        start_position + position
        for start_position, position in zip(start_positions, positions)
    ])

    locdata_dict = {
        'position_' + label: position_values
        for _, position_values, label in zip(ranges, new_positions.T, ('x', 'y', 'z'))
    }

    locdata_dict.update(frame=np.tile(range(len(times)), n_walks))

    locdata = LocData.from_dataframe(dataframe=pd.DataFrame(locdata_dict))

    # metadata
    locdata.meta.source = metadata_pb2.SIMULATION
    del locdata.meta.history[:]
    locdata.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))

    return locdata


def resample(locdata, n_samples=10, seed=None):
    """
    Resample locdata according to localization uncertainty. Per localization *n_samples* new localizations
    are simulated normally distributed around the localization coordinates with a standard deviation set to the
    uncertainty in each dimension.
    The resulting LocData object carries new localizations with the following properties: 'position_x',
    'position_y'[, 'position_z'], 'origin_index'

    Parameters
    ----------
    locdata : LocData
        Localization data to be resampled
    n_samples : int
        The number of localizations generated for each original localization.
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    locdata : LocData
        New localization data with simulated coordinates.
    """
    rng = np.random.default_rng(seed)

    # generate dataframe
    list_ = []
    for i in range(len(locdata)):
        new_d = {}
        new_d.update({'origin_index': np.full(n_samples, i)})
        x_values = rng.normal(loc=locdata.data.iloc[i]['position_x'],
                              scale=locdata.data.iloc[i]['uncertainty_x'],
                              size=n_samples
                              )
        new_d.update({'position_x': x_values})

        y_values = rng.normal(loc=locdata.data.iloc[i]['position_y'],
                              scale=locdata.data.iloc[i]['uncertainty_y'],
                              size=n_samples
                              )
        new_d.update({'position_y': y_values})

        try:
            z_values = rng.normal(loc=locdata.data.iloc[i]['position_z'],
                                  scale=locdata.data.iloc[i]['uncertainty_z'],
                                  size=n_samples
                                  )
            new_d.update({'position_z': z_values})
        except KeyError:
            pass

        list_.append(pd.DataFrame(new_d))

    dataframe = pd.concat(list_, ignore_index=True)

    # metadata
    meta_ = metadata_pb2.Metadata()
    meta_.CopyFrom(locdata.meta)
    try:
        meta_.ClearField("identifier")
    except ValueError:
        pass

    try:
        meta_.ClearField("element_count")
    except ValueError:
        pass

    try:
        meta_.ClearField("frame_count")
    except ValueError:
        pass

    meta_.modification_time.GetCurrentTime()
    meta_.state = metadata_pb2.MODIFIED
    meta_.ancestor_identifiers.append(locdata.meta.identifier)
    meta_.history.add(name='resample',
                      parameter='locdata={}, n_samples={}'.format(locdata, n_samples))

    # instantiate
    new_locdata = LocData.from_dataframe(dataframe=dataframe, meta=meta_)

    return new_locdata


def _random_poisson_repetitions(n_samples, lam, seed=None):
    """
    Return numpy.ndarray of sorted integers with each integer i being repeated n(i) times
    where n(i) is drawn from a Poisson distribution with mean `lam`.

    Parameters
    ----------
    n_samples : int
        number of elements to be returned
    lam : float
        mean of the Poisson distribution (lambda)
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    numpy.ndarray with shape (n_samples,)
        The generated sequence of integers.
    """
    rng = np.random.default_rng(seed)

    frames = np.ones(n_samples, dtype=int)
    n_random_numbers = n_samples if lam > 0 else int(n_samples / lam)
    position = 0
    current_number = 0
    while position < n_samples:
        repeats = rng.poisson(lam=lam, size=n_random_numbers)
        for repeat in repeats:
            try:
                frames[position:position + repeat] = current_number
            except ValueError:
                frames[position:] = current_number
                break
            position += repeat
            current_number += 1
    return frames


def simulate_frame_numbers(n_samples, lam, seed=None):
    """
    Simulate Poisson-distributed frame numbers for a list of localizations.

    Return numpy.ndarray of sorted integers with each integer i being repeated n(i) times
    where n(i) is drawn from a Poisson distribution with mean `lam`.

    Use the following to add frame numbers to a given LocData object::

        frames = simulate_frame_numbers(n_samples=len(locdata), lam=5)
        locdata.dataframe = locdata.dataframe.assign(frame = frames)

    Parameters
    ----------
    n_samples : int
        number of elements to be returned
    lam : float
        mean of the Poisson distribution (lambda)
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        random number generation seed

    Returns
    -------
    numpy.ndarray with shape (n_samples,)
        The generated sequence of integers.
    """
    return _random_poisson_repetitions(n_samples, lam, seed=seed)



from copy import deepcopy

import pytest
import pandas as pd
import matplotlib.pyplot as plt

from locan import LocData, LocalizationsPerFrame
from locan.analysis.localizations_per_frame import _Results, _localizations_per_frame, _DistributionFits


def test__localizations_per_frame(caplog, locdata_rapidSTORM_2d):
    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d)
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index[-1] == 47
    assert series.name == "n_localizations"

    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d, norm=2, time_delta=None)
    assert len(series) == 48
    assert series.iloc[0] == 11
    assert series.index[-1] == 47
    assert series.name == "n_localizations / 2"

    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d, norm=2)
    assert len(series) == 48
    assert series.iloc[0] == 11
    assert series.index[-1] == 47

    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d, time_delta=10)
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index.seconds[-1] == 47 * 10
    assert series.name == "n_localizations / s"

    series = _localizations_per_frame(locdata=range(10))
    assert len(series) == 10
    assert series.iloc[0] == 1
    assert series.index[-1] == 9
    assert series.name == "n_localizations"

    series = _localizations_per_frame(locdata=locdata_rapidSTORM_2d, time_delta=10, resample="10s")
    assert len(series) == 48
    assert series.iloc[0] == 22
    assert series.index.seconds[-1] == 47 * 10
    assert series.name == "n_localizations / s"

    with pytest.raises(TypeError):
        _localizations_per_frame(locdata=range(10), resample="2s")

    assert caplog.record_tuples == [('locan.analysis.localizations_per_frame', 30,
                                     'integration_time not available in locdata.meta - frames used instead.'),
                                    ('locan.analysis.localizations_per_frame', 30,
                                     'integration_time not available in locdata.meta - frames used instead.'),
                                    ('locan.analysis.localizations_per_frame', 30,
                                     'integration_time not available in locdata.meta - frames used instead.'),
                                    ('locan.analysis.localizations_per_frame', 30,
                                     'integration_time not available in locdata.meta - frames used instead.')
                                    ]


class TestLocalizationsPerFrame:

    def test_init(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame(meta={'comment': 'this is an example'})
        assert str(lpf) == "LocalizationsPerFrame(norm=None, time_delta=integration_time, resample=None)"
        assert lpf.results is None
        assert lpf.meta.comment == 'this is an example'

    def test_empty_locdata(self, caplog):
        lpf = LocalizationsPerFrame().compute(LocData())
        lpf.fit_distributions()
        lpf.plot()
        lpf.hist()
        # plt.show()

        assert caplog.record_tuples == [('locan.analysis.localizations_per_frame', 30, 'Locdata is empty.'),
                                        ('locan.analysis.localizations_per_frame', 30, 'No results available to fit.')]

        plt.close('all')

    def test_compute(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame().compute(locdata=locdata_rapidSTORM_2d)
        assert (isinstance(lpf.results, _Results))
        assert(isinstance(lpf.results.time_series, pd.Series))

        # print(lpf.results.time_series.head())
        assert len(lpf.results.time_series) == 48
        assert lpf.results.time_series.iloc[0] == 22
        assert lpf.results.time_series.index[-1] == 47

        assert lpf.results.accumulation_time() == 22
        assert lpf.results.accumulation_time(fraction=0.8) == 38

        lpf.plot()
        lpf.plot(window=10)
        lpf.hist()
        # plt.show()

        lpf.fit_distributions(floc=0)
        assert(0 in lpf.distribution_statistics.parameter_dict().values())

        plt.close('all')

    def test_resample(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame(norm='localization_density_bb', time_delta=1, resample="2s")\
            .compute(locdata=locdata_rapidSTORM_2d)
        assert(lpf.results.time_series.name == 'n_localizations / localization_density_bb / s')

        lpf.plot()
        lpf.plot(window=10)
        lpf.hist()
        # plt.show()

        plt.close('all')

    def test_norm(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame(norm=5).compute(locdata=locdata_rapidSTORM_2d)
        assert(lpf.results.time_series.name == 'n_localizations / 5')

        locdata_rapidSTORM_2d_ = deepcopy(locdata_rapidSTORM_2d)
        od = locdata_rapidSTORM_2d_.meta.experiment.setups.add().optical_units.add()
        od.detection.camera.integration_time.FromMilliseconds(10)
        lpf = LocalizationsPerFrame(norm=None).compute(locdata=locdata_rapidSTORM_2d_)
        assert(lpf.results.time_series.name == 'n_localizations / s')


@pytest.mark.visual
class TestLocalizationsPerFrameVisual:

    @pytest.fixture()
    def lpf(self, locdata_rapidSTORM_2d):
     return LocalizationsPerFrame().compute(locdata=locdata_rapidSTORM_2d)

    def test_plot(self, lpf):
        lpf.plot()
        plt.show()

        lpf.plot(cumulative=True)
        plt.show()

        lpf.plot(cumulative=True, normalize=True)
        plt.show()

        plt.close('all')

    def test_hist(self, lpf):
        lpf.hist(density=False)
        plt.show()

        lpf.hist(density=True, cumulative=True)
        plt.show()

        plt.close('all')


class TestDistributionFits:

    def test_fit(self, locdata_rapidSTORM_2d):
        lpf = LocalizationsPerFrame().compute(locdata=locdata_rapidSTORM_2d)
        distribution_statistics = _DistributionFits(lpf)
        # print(distribution_statistics)
        assert distribution_statistics.parameter_dict() == {}

        distribution_statistics.fit()
        assert(list(distribution_statistics.parameter_dict().keys()) ==
               ['n_localizations_center', 'n_localizations_sigma'])

        distribution_statistics.plot()
        # plt.show()

        plt.close('all')



"""
Building an analysis pipeline.

Pipeline refers to sequential analysis steps that are applied to a single LocData object.
An analysis pipeline here includes true piped analysis, where a preliminary result serves as input to the next analysis
step, but also workflows that provide different results in parallel.

A batch process is a procedure for running a pipeline over multiple LocData objects while collecting and combing
results.

This module provides a class `Pipeline` to combine the analysis procedure, parameters and results
in a single pickleable object.
"""
import inspect
import logging

from locan.analysis.analysis_base import _Analysis


__all__ = ['Pipeline']

logger = logging.getLogger(__name__)


class Pipeline(_Analysis):
    """
    The base class for a specialized analysis pipeline to be used on LocData objects.

    The custom analysis routine has to be added by implementing the method `computation(self, **kwargs)`.
    Keyword arguments must include the locdata reference and optional parameters.

    Results are provided as customized attributes.
    We suggest abbreviated standard names for the most common procedures such as:

    * lp - Localization Precision
    * lprop - Localization Property
    * lpf - Localizations per Frame
    * rhf - Ripley H function
    * clust - locdata with clustered elements

    Parameters
    ----------
    computation : callable
        A function `computation(self, **kwargs)` specifying the analysis procedure.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    kwargs : dict
        Locdata reference and optional parameters passed to `computation(self, **kwargs)`.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    computation : callable
        A function `computation(self, **kwargs)` specifying the analysis procedure.
    kwargs : dict
        All parameters including the locdata reference that are passed to `computation(self, **kwargs)`.

    Note
    ----
    The class variable `Pipeline.count` is only incremented in a single process. In multiprocessing `Pipeline.count` and
    `Pipeline.meta.identifier` (which is set using `count`) cannot be used to identify distinct Pipeline objects.

    Note
    ----
    For the Pipeline object to be pickleable attention has to be paid to the :func:`computation` method.
    With multiprocessing it will have to be re-injected for each Pipeline object by `pipeline.computation = computation`
    after computation and before pickling.
    """
    def __init__(self, computation, meta=None, **kwargs):
        super().__init__(meta=meta, computation=computation, **kwargs)

        if not callable(computation):
            raise TypeError('A callable function `computation(self, locdata, **kwargs)` '
                            'must be passed as first argument.')
        self.computation = computation
        self.kwargs = kwargs

    def __bool__(self):
        return True

    def compute(self):
        """ Run the analysis procedure. All parameters must be given upon Pipeline instantiation."""
        return self.computation(self, **self.kwargs)

    def save_computation(self, path):
        """
        Save the analysis procedure (i.e. the computation() method) as human readable text.

        Parameters
        ----------
        path : str, bytes, os.PathLike
            Path and file name for saving the text file.
        """
        with open(path, 'w') as handle:
            handle.write('Analysis Pipeline: {}\n\n'.format(self.__class__.__name__))
            handle.write(inspect.getsource(self.computation))

    def computation_as_string(self):
        """
        Return the analysis procedure (i.e. the computation() method) as string.
        """
        return inspect.getsource(self.computation)


def computation_test(self, locdata=None, parameter='test'):
    """ A pipeline definition for testing."""
    self.locdata = locdata
    something = 'changed_value'
    logger.debug(f'something has a : {something}')
    self.test = parameter
    logger.info(f'computation finished for locdata: {locdata}')

    try:
        raise NotImplementedError
    except NotImplementedError:
        logger.warning(f'An exception occurred for locdata: {locdata}')

    return self



"""

Provide standard formats.

"""
import time


__all__ = []


def _time_string(time_value):
    """
    Convert 'time_value' (typically timestamp from Unix epoch) to the local time
    and return as "%Y-%m-%d %H:%M:%S %z" formatted string.

    This time format is used in metadata of LocData and Analysis classes.

    Parameters
    ----------
    time_value : float
        Return value from :func:`time.time`

    Returns
    -------
    str
    """
    return time.strftime("%Y-%m-%d %H:%M:%S %z", time.localtime(time_value))


"""

File input/output for localization data in ASDF files

"""
from __future__ import annotations
import logging

import pandas as pd
from asdf import AsdfFile
from asdf import open as asdf_open
from google.protobuf import json_format

from locan.data.locdata import LocData
from locan.locan_io.locdata.utilities import convert_property_names


__all__ = ['save_asdf', 'load_asdf_file']

logger = logging.getLogger(__name__)


def save_asdf(locdata, path):
    """
    Save LocData attributes in an asdf file.

    In the Advanced Scientific Data Format (ASDF) file format we store metadata, properties and column names as human-
    readable yaml header. The data is stored as binary numpy.ndarray.

    Note
    ----
    Only selected LocData attributes are saved.
    Currently these are: 'data', 'columns', 'properties', 'meta'.

    Parameters
    ----------
    locdata : LocData
        The LocData object to be saved.
    path : str, bytes, os.PathLike, file-like
        File path including file name to save to.
    """
    # Prepare tree
    meta_json = json_format.MessageToJson(locdata.meta, including_default_value_fields=False)
    tree = {
        'data': locdata.data.values,
        'columns': list(locdata.data),
        'properties': locdata.properties,
        'meta': meta_json
    }

    # Create the ASDF file object from tree
    af = AsdfFile(tree)

    # Write the data to a new file
    af.write_to(path)


def load_asdf_file(path, nrows=None):
    """
    Load data from ASDF localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a rapidSTORM file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with asdf_open(path) as af:
        new_df = pd.DataFrame({k: af.tree['data'][slice(nrows), n] for n, k in enumerate(af.tree['columns'])})

        column_keys = convert_property_names(properties=new_df.columns.tolist(), property_mapping=None)
        mapper = {key: value for key, value in zip(new_df.columns, column_keys)}
        new_df = new_df.rename(columns=mapper)

        locdata = LocData(dataframe=new_df)
        locdata.meta = json_format.Parse(af.tree['meta'], locdata.meta, ignore_unknown_fields=True)
    return locdata



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: locan/analysis/metadata_analysis.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&locan/analysis/metadata_analysis.proto\x12\x0elocan.analysis\x1a\x1fgoogle/protobuf/timestamp.proto\"3\n\x10\x41nalysis_routine\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tparameter\x18\x02 \x01(\t\"\xd7\x02\n\tAMetadata\x12\x12\n\nidentifier\x18\x01 \x01(\t\x12\x0f\n\x07\x63omment\x18\x02 \x01(\t\x12\x30\n\x06method\x18\x05 \x01(\x0b\x32 .locan.analysis.Analysis_routine\x12/\n\x03map\x18\x07 \x03(\x0b\x32\".locan.analysis.AMetadata.MapEntry\x12\x31\n\rcreation_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x35\n\x11modification_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01J\x04\x08\x03\x10\x04J\x04\x08\x04\x10\x05R\rcreation_dateR\x11modification_date')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'locan.analysis.metadata_analysis_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _AMETADATA_MAPENTRY._options = None
  _AMETADATA_MAPENTRY._serialized_options = b'8\001'
  _ANALYSIS_ROUTINE._serialized_start=91
  _ANALYSIS_ROUTINE._serialized_end=142
  _AMETADATA._serialized_start=145
  _AMETADATA._serialized_end=488
  _AMETADATA_MAPENTRY._serialized_start=400
  _AMETADATA_MAPENTRY._serialized_end=442
# @@protoc_insertion_point(module_scope)



import matplotlib.colors as mcolors

from locan.dependencies import HAS_DEPENDENCY
from locan import PropertyDescription, PropertyKey, PROPERTY_KEYS, \
    FileType, HullType, RenderEngine, ColorMaps, \
    RAPIDSTORM_KEYS, ELYRA_KEYS, THUNDERSTORM_KEYS, NANOIMAGER_KEYS, SMLM_KEYS, DECODE_KEYS, SMAP_KEYS


def test_PropertyDescription():
    prop = PropertyDescription(name="prop_name", type='integer', unit_SI="m", unit="nm", description="something")
    assert repr(prop) == "PropertyDescription(name='prop_name', type='integer', unit_SI='m', unit='nm', " \
                         "description='something')"
    assert prop.name == "prop_name"


def test_PropertyKey():
    assert "index" in PropertyKey.__members__
    for element in PropertyKey:
        assert element.name == element.value.name

    assert all(key in PropertyKey.coordinate_labels()
               for key in [PropertyKey.position_x, PropertyKey.position_y, PropertyKey.position_z])

    PropertyKey.position_x.value.unit = "nm"
    assert PropertyKey.position_x.value.unit == "nm"

    string_ = PropertyKey.summary()
    # print(string_)
    assert string_[:5] == "index"


def test_PROPERTY_KEYS():
    assert all(key in PropertyKey._member_names_ for key in PROPERTY_KEYS.keys())


def test_HullType():
    values = [item.value for item in HullType]
    assert all(value in values for value in ['bounding_box', 'convex_hull', 'oriented_bounding_box', 'alpha_shape'])


def test_FileType():
    from locan.data.metadata_pb2 import File_type
    for ft_enum, ft_pb in zip(FileType, File_type.items()):
        assert ft_enum.name == ft_pb[0]
        assert ft_enum.value == ft_pb[1]


def test_RenderEngine():
    if not HAS_DEPENDENCY["napari"]:
        assert all(key in RenderEngine._member_names_ for key in ['MPL'])
    else:
        assert all(key in RenderEngine._member_names_ for key in ['MPL', 'NAPARI'])


def test_ColorMaps():
    assert all([isinstance(item.value, mcolors.Colormap) for item in ColorMaps])


def test_keys_are_mapped_on_valid_property_keys():
    assert all(PropertyKey[value] for value in RAPIDSTORM_KEYS.values())
    assert all(PropertyKey[value] for value in ELYRA_KEYS.values())
    assert all(PropertyKey[value] for value in THUNDERSTORM_KEYS.values())
    assert all(PropertyKey[value] for value in NANOIMAGER_KEYS.values())
    assert all(PropertyKey[value] for value in SMLM_KEYS.values())
    assert all(PropertyKey[value] for value in DECODE_KEYS.values())
    assert all(PropertyKey[value] for value in SMAP_KEYS.values())



"""

Hull objects of localization data.

This module computes specific hulls for the bounding box, convex hull and oriented bounding box
and related properties for LocData objects.

"""
import numpy as np
import scipy.spatial as spat
from shapely.geometry import MultiPoint as shMultiPoint

from locan.data.region import Rectangle, Polygon


__all__ = ['BoundingBox', 'ConvexHull', 'OrientedBoundingBox']


class BoundingBox:
    """
    Class with bounding box computed using numpy operations.

    Parameters
    ----------
    points : numpy.ndarray of double, shape (npoints, ndim)
        Coordinates of input points.

    Attributes
    ----------
    hull : array of arrays
        Array of point coordinates that represent [[min_coordinates], [max_coordinates]].
    dimension : int
        Spatial dimension of hull
    vertices : array of coordinate tuples
        Coordinates of points that make up the hull.
    width : array of float
        Array with differences between max and min for each coordinate.
    region_measure : float
        Hull measure, i.e. area or volume
    subregion_measure : float
        Measure of the sub-dimensional region, i.e. circumference or surface
    region : RoiRegion
        Convert the hull to a RoiRegion object.
    """

    def __init__(self, points):
        if np.size(points) == 0:
            self.dimension = 0
            self.hull = np.array([])
            self.width = np.zeros(self.dimension)
            self.region_measure = 0
            self.subregion_measure = 0
        elif len(points) < 2:
            self.dimension = np.shape(points)[1]
            self.hull = np.array([])
            self.width = np.zeros(self.dimension)
            self.region_measure = 0
            self.subregion_measure = 0
        else:
            self.dimension = np.shape(points)[1]
            self.hull = np.array([np.min(points, axis=0), np.max(points, axis=0)])
            self.width = np.diff(self.hull, axis=0).flatten()
            self.region_measure = np.prod(self.width)
            self.subregion_measure = np.sum(self.width)*2

    @property
    def vertices(self):
        return self.hull.T

    @property
    def region(self):
        if self.dimension == 2:
            region_ = Rectangle(self.hull[0], self.width[0], self.width[1], 0)
            # region_ = RoiRegion(region_type='rectangle', region_specs=(self.hull[0], self.width[0], self.width[1], 0))
        else:
            raise NotImplementedError
        return region_


class _ConvexHullScipy:
    """
    Class with convex hull computed using the scipy.spatial.ConvexHull method.

    Parameters
    ----------
    points : numpy.ndarray of double, shape (npoints, ndim)
        Coordinates of input points.

    Attributes
    ----------
    hull : Hull object
        hull object from the corresponding algorithm
    dimension : int
        spatial dimension of hull
    vertices : array of coordinate tuples
        Coordinates of points that make up the hull.
    vertex_indices : indices for points
        indices identifying a polygon of all points that make up the hull
    points_on_boundary : int
        absolute number of points that are part of the convex hull.
    points_on_boundary_rel : float
        The number of points on the hull relative to all input points
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : Region
        Convert the hull to a Region object.
    """

    def __init__(self, points):
        if len(points) < 6:
            unique_points = np.array(list(set(tuple(point) for point in points)))
            if len(unique_points) < 3:
                raise TypeError('Convex_hull needs at least 3 different points as input.')

        self.dimension = np.shape(points)[1]
        self.hull = spat.ConvexHull(points)
        self.vertex_indices = self.hull.vertices
        self.points_on_boundary = len(self.vertex_indices)
        self.points_on_boundary_rel = self.points_on_boundary / len(points)
        self.region_measure = self.hull.volume
        self.subregion_measure = self.hull.area

    @property
    def vertices(self):
        return self.hull.points[self.hull.vertices]

    @property
    def region(self):
        if self.dimension > 2:
            raise NotImplementedError('Region for 3D data has not yet been implemented.')
        else:
            # closed_vertices = np.append(self.vertices, [self.vertices[0]], axis=0)
            # region_ = RoiRegion(region_type='polygon', region_specs=closed_vertices)
            return Polygon(self.vertices)


class _ConvexHullShapely:
    """
    Class with convex hull computed using the scipy.spatial.ConvexHull method.

    Parameters
    ----------
    points : numpy.ndarray of double, shape (npoints, ndim)
        Coordinates of input points.

    Attributes
    ----------
    hull : Hull object
        Polygon object from the .convex_hull method
    dimension : int
        Spatial dimension of hull
    vertices : array of coordinate tuples
        Coordinates of points that make up the hull.
    vertex_indices : indices for points
        indices identifying a polygon of all points that make up the hull
    points_on_boundary : int
        The absolute number of points on the hull
    points_on_boundary_rel : int
        The number of points on the hull relative to all input points
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : RoiRegion
        Convert the hull to a RoiRegion object.
    """
    def __init__(self, points):
        if len(points) < 6:
            unique_points = np.array(list(set(tuple(point) for point in points)))
            if len(unique_points) < 3:
                raise TypeError('Convex_hull needs at least 3 different points as input.')

        self.dimension = np.shape(points)[1]
        if self.dimension >= 3:
            raise TypeError('ConvexHullShapely only takes 1 or 2-dimensional points as input.')

        self.hull = shMultiPoint(points).convex_hull
        # todo: set vertex_indices
        # self.vertex_indices = None
        self.points_on_boundary = len(self.hull.exterior.coords)-1  # the first point is repeated in exterior.coords
        self.points_on_boundary_rel = self.points_on_boundary / len(points)
        self.region_measure = self.hull.area
        self.subregion_measure = self.hull.length

    @property
    def vertices(self):
        return np.array(self.hull.exterior.coords)[:-1]

    @property
    def region(self):
        if self.dimension > 2:
            raise NotImplementedError('Region for 3D data has not yet been implemented.')
        else:
            #  closed_vertices = np.append(self.vertices, [self.vertices[0]], axis=0)
            # region_ = RoiRegion(region_type='polygon', region_specs=closed_vertices)
            return Polygon(self.vertices)


class ConvexHull:
    """
    Class with convex hull of localization data.

    Parameters
    ----------
    points : numpy.ndarray of double, shape (npoints, ndim)
        Coordinates of input points.
    method : string
        Specific class to compute the convex hull and attributes. One of 'scipy', 'shapely'.

    Attributes
    ----------
    method : string
        Specific class to compute the convex hull and attributes. One of 'scipy', 'shapely'.
    hull : Hull object
        Polygon object from the .convex_hull method
    dimension : int
        Spatial dimension of hull
    vertices : array of coordinate tuples
        Coordinates of points that make up the hull.
    vertex_indices : indices for points
        indices identifying a polygon of all points that make up the hull
    points_on_boundary : int
        The absolute number of points on the hull
    points_on_boundary_rel : int
        The number of points on the hull relative to all input points
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : Region
        Convert the hull to a Region object.
    """

    def __init__(self, points, method='scipy'):
        self.method = method
        if method == 'scipy':
            self._special_class = _ConvexHullScipy(points)
        elif method == 'shapely':
            self._special_class = _ConvexHullShapely(points)
        else:
            raise ValueError(f'The provided method {method} is not available.')

    def __getattr__(self, attr):
        if attr.startswith('__') and attr.endswith('__'):  # this is needed to enable pickling
            raise AttributeError

        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._special_class, attr)


class OrientedBoundingBox:
    """
    Class with oriented bounding box computed using the shapely minimum_rotated_rectangle method.

    Parameters
    ----------
    points : numpy.ndarray of double, shape (npoints, ndim)
        Coordinates of input points.

    Attributes
    ----------
    hull : hull object
        Polygon object from the minimum_rotated_rectangle method
    dimension : int
        Spatial dimension of hull
    vertices : array of coordinate tuples
        Coordinates of points that make up the hull.
    width : array of float
        Array with lengths of box edges.
    region_measure : float
        hull measure, i.e. area or volume
    subregion_measure : float
        measure of the sub-dimensional region, i.e. circumference or surface
    region : Region
        Convert the hull to a Region object.
    angle : float
        Orientation defined as angle (in degrees) between the vector from first to last point and x-axis.
    
    """
    def __init__(self, points):
        self.dimension = np.shape(points)[1]

        if self.dimension >= 3:
            raise TypeError('OrientedBoundingBox only takes 1 or 2-dimensional points as input.')

        if len(points) < 3:
            self.hull = np.array([])
            self.width = np.zeros(self.dimension)
            self.region_measure = 0
            self.subregion_measure = 0
            self.angle = np.nan
            self.elongation = np.nan
        else:
            self.hull = shMultiPoint(points).minimum_rotated_rectangle
            difference = np.diff(self.vertices[0:3], axis=0)
            self.width = np.array([np.linalg.norm(difference[0]), np.linalg.norm(difference[1])])
            self.region_measure = self.hull.area
            self.subregion_measure = self.hull.length
            self.angle = np.degrees(np.arctan2(difference[0][1], difference[0][0]))
            # numpy.arctan2(y, x) takes reversed x, y arguments.
            self.elongation = 1 - np.divide(*sorted(self.width))

    @property
    def vertices(self):
        return np.array(self.hull.exterior.coords)

    @property
    def region(self):
        if self.dimension == 2:
            # region_ = RoiRegion(region_type='rectangle',
            #                     region_specs=(self.vertices[0], self.width[0], self.width[1], self.angle))
            return Rectangle(self.vertices[0], self.width[0], self.width[1], self.angle)
        else:
            raise NotImplementedError



"""
.. currentmodule:: locan

Locan consists of the following modules:

.. autosummary::
   :toctree: generated/

   analysis
   constants
   configuration
   data
   datasets
   dependencies
   gui
   locan_io
   render
   scripts
   simulation
   utils
   tests
"""
from pathlib import Path
import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

try:
    from locan._version import version as __version__
except ImportError:
    __version__ = "not-installed"


# Root directory for path operations.
ROOT_DIR = Path(__file__).parent

# Identifier for LocData objects that is reset for each locan session and incremented with each LocData instantiation.
locdata_id = 0


from locan.dependencies import *  # has to be imported before any others due to pyqt5/sidepy2 issues.
from locan.constants import *
from locan.configuration import *
from locan.analysis import *
from locan.data import *
from locan.datasets import *
from locan.gui import *
from locan.locan_io import *
from locan.render import *
from locan.simulation import *
from locan.utils import *
from locan.tests import *


__all__ = ['__version__', 'ROOT_DIR']
__all__.extend(dependencies.__all__)
__all__.extend(constants.__all__)
__all__.extend(configuration.__all__)
__all__.extend(analysis.__all__)
__all__.extend(data.__all__)
__all__.extend(datasets.__all__)
__all__.extend(gui.__all__)
__all__.extend(locan_io.__all__)
__all__.extend(render.__all__)
__all__.extend(simulation.__all__)
__all__.extend(utils.__all__)
__all__.extend(tests.__all__)



"""

Utility methods for clustering locdata.

"""
import sys
import itertools
from inspect import signature

from locan.data.locdata import LocData


__all__ = ['serial_clustering']


def serial_clustering(locdata, algorithm, parameter_lists, **kwargs):
    """
    Run and analyse a series of clustering processes to identify optimal parameters.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    algorithm : callable
        The locan clustering algorithm to use on locdata.
    parameter_lists : dict
        A dictionary with all parameter lists that are to be iterated. The keys should be identical to parameter names
        of the used algorithm.
    kwargs : dict
        Optional keyword arguments that are passed to the algorithm.

    Returns
    -------
    Tuple of LocData
        The first element is a LocData object with a selection of all localizations that are defined as noise.
        If noise is false this element will be None.
        The second element is a new LocData instance assembling all generated selections (i.e. localization cluster).

    """
    parameter = locals()

    keys = parameter_lists.keys()
    products = list(itertools.product(*parameter_lists.values()))
    products_transposed = zip(*products)
    dictionary = {key: values for key, values in zip(keys, products_transposed)}
    results = []
    for l in products:
        iterated_arguments = {k: v for k, v in zip(keys, l)}
        results.append(algorithm(locdata, **iterated_arguments, **kwargs))

    noise_locdata = [result[0] for result in results]
    noise_collection = LocData.from_collection(noise_locdata)
    noise_collection.dataframe = noise_collection.dataframe.assign(**dictionary)
    collection_locdata = [result[1] for result in results]
    collection = LocData.from_collection(collection_locdata)
    collection.dataframe = collection.dataframe.assign(**dictionary)

    # metadata for noise_collection
    del noise_collection.meta.history[:]
    noise_collection.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))

    # metadata for collection
    del collection.meta.history[:]
    collection.meta.history.add(name=sys._getframe().f_code.co_name, parameter=str(parameter))

    return noise_collection, collection



import pytest
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt  # this import is needed for visual inspection

from locan import LocData, simulate_uniform
from locan import NearestNeighborDistances  # from locan.analysis.nearest_neighbor
from locan.analysis.nearest_neighbor import pdf_nnDistances_csr_2D, pdf_nnDistances_csr_3D, \
    NNDistances_csr_2d, NNDistances_csr_3d, _DistributionFits


# fixtures

@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        'position_x': [0, 0, 1, 4, 5],
        'position_y': [0, 1, 3, 4, 1],
        }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def other_locdata_simple():
    locdata_dict = {
        'position_x': [10, 11],
        'position_y': [10, 11],
        }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


# tests

def test_pdf_nnDistances_csr_2D():
    pdf_values = pdf_nnDistances_csr_2D(np.array([0, 1, 2]), density=1)
    assert pdf_values == pytest.approx([0, 2.71521056e-001, 4.38232365e-05])


def test_pdf_nnDistances_csr_3D():
    pdf_values = pdf_nnDistances_csr_3D(np.array([0, 1, 2]), density=1)
    assert pdf_values == pytest.approx([0, 1.90564233e-01, 1.40579528e-13])


def test_NNDistances_csr_2d():
    dist = NNDistances_csr_2d()
    assert dist.shapes == 'density'
    assert dist.pdf(x=np.array([0, 1, 2]), density=1) == pytest.approx([0, 2.71521056e-01, 4.38232365e-05])


def test_NNDistances_csr_3d():
    dist = NNDistances_csr_3d()
    assert dist.shapes == 'density'
    assert dist.pdf(x=np.array([0, 1, 2]), density=1) == pytest.approx([0, 1.90564233e-01, 1.40579528e-13])


def test_Nearest_neighbor_distances_empty(caplog):
    nn_1 = NearestNeighborDistances().compute(LocData())
    nn_1.hist()
    assert caplog.record_tuples == [('locan.analysis.nearest_neighbor', 30, 'Locdata is empty.')]


def test_DistributionFits(locdata_simple):
    nn_1 = NearestNeighborDistances()
    assert repr(nn_1) == "NearestNeighborDistances(k=1)"
    nn_1.compute(locdata_simple)
    assert nn_1.dimension == 2
    assert nn_1.localization_density == 0.25
    # print(nn_1.results)
    assert len(nn_1.results) == 5
    assert all(nn_1.results.columns == ['nn_distance', 'nn_index'])
    assert nn_1.results['nn_index'].iloc[0] == 1
    assert nn_1.distribution_statistics is None
    # nn_1.hist()

    ds = _DistributionFits(nn_1)
    ds.fit()
    assert ds.parameter_dict() == {'density': pytest.approx(0.058984374999999166), 'loc': 0, 'scale': 1}
    ds.fit(with_constraints=False)
    assert ds.parameter_dict() == pytest.approx({'density': 0.8900132393760514,
                                                 'loc': 0.35972770464810505,
                                                 'scale': 3.348388757318209})
    # ds.plot()

    nn_1.hist(fit=True)
    assert nn_1.distribution_statistics.parameter_dict() == \
           {'density': pytest.approx(0.058984374999999166), 'loc': 0, 'scale': 1}
    # plt.show()

    plt.close('all')


def test_DistributionFits_k2(locdata_simple):
    nn_1 = NearestNeighborDistances(k=2)
    assert repr(nn_1) == "NearestNeighborDistances(k=2)"
    nn_1.compute(locdata_simple)
    assert nn_1.dimension == 2
    assert nn_1.localization_density == 0.25
    # print(nn_1.results)
    assert len(nn_1.results) == 5
    assert all(nn_1.results.columns == ['nn_distance', 'nn_index'])
    assert nn_1.results['nn_index'].iloc[0] == 2
    assert nn_1.distribution_statistics is None
    nn_1.fit_distributions()
    assert nn_1.distribution_statistics.parameter_dict() == \
           {'density': pytest.approx(0.028906249999999137), 'loc': 0, 'scale': 1}

    nn_1.hist(fit=True)
    # plt.show()

    plt.close('all')


def test_Nearest_neighbor_distances(locdata_simple, other_locdata_simple):
    nn_2 = NearestNeighborDistances().compute(locdata_simple, other_locdata=other_locdata_simple)
    # print(nn_2.results)
    assert nn_2.results['nn_distance'].iloc[0] == pytest.approx(14.142135623730951)


def test_NearestNeighborDistances_3d(locdata_3d):
    # print(locdata_3d.data)
    nn_1 = NearestNeighborDistances().compute(locdata_3d)
    assert repr(nn_1) == "NearestNeighborDistances(k=1)"
    assert nn_1.dimension == 3
    assert nn_1.localization_density == 0.075
    # print(nn_1.results)
    assert len(nn_1.results) == 6
    assert all(nn_1.results.columns == ['nn_distance', 'nn_index'])
    assert nn_1.results['nn_index'].iloc[0] == 4
    assert nn_1.distribution_statistics is None
    # nn_1.hist()

    ds = _DistributionFits(nn_1)
    ds.fit()
    assert ds.parameter_dict() == {'density': pytest.approx(0.007666015624999118), 'loc': 0, 'scale': 1}
    ds.fit(with_constraints=False)
    assert ds.parameter_dict() == pytest.approx({'density': 1.0047585928889682,
                                                 'loc': 2.4220565244971968,
                                                 'scale': 1.3178818824051617})
    # ds.plot()

    nn_1.hist(fit=True)
    assert nn_1.distribution_statistics.parameter_dict() == \
           {'density': pytest.approx(0.007666015624999118), 'loc': 0, 'scale': 1}
    # plt.show()

    plt.close('all')


def test_NearestNeighborDistances_1d(caplog, locdata_1d):
    nn_1 = NearestNeighborDistances().compute(locdata_1d)
    assert repr(nn_1) == "NearestNeighborDistances(k=1)"
    assert nn_1.dimension == 1
    assert nn_1.localization_density == 1.5
    # print(nn_1.results)
    assert len(nn_1.results) == 6
    assert all(nn_1.results.columns == ['nn_distance', 'nn_index'])
    assert nn_1.results['nn_index'].iloc[0] == 1
    assert nn_1.distribution_statistics is None
    # nn_1.hist()

    ds = _DistributionFits(nn_1)
    ds.fit()
    assert caplog.record_tuples == [('locan.analysis.nearest_neighbor', 30, 'No fit model for 1 dimensions available.')]
    assert ds.parameter_dict() == {}
    ds.plot()

    nn_1.hist(fit=True)
    # plt.show()

    plt.close('all')


@pytest.mark.visual
def test_NearestNeighborDistances_2d_random():
    locdata = simulate_uniform(n_samples=10_000, region=((0, 1), (0, 1)))
    nn_1 = NearestNeighborDistances().compute(locdata)
    assert nn_1.localization_density == locdata.properties['localization_density_bb']
    assert len(nn_1.results) == 10_000
    nn_1.hist(fit=True)
    plt.show()


@pytest.mark.visual
def test_NearestNeighborDistances_3d_random():
    locdata = simulate_uniform(n_samples=10_000, region=((0, 1), (0, 1), (0, 1)))
    nn_1 = NearestNeighborDistances().compute(locdata)
    assert nn_1.localization_density == locdata.properties['localization_density_bb']
    assert len(nn_1.results) == 10_000
    nn_1.hist(fit=True)
    plt.show()



from abc import ABC, abstractmethod, abstractproperty, abstractclassmethod


class LocDataBase(ABC):

    @abstractproperty
    def data(self):
        return self.dataframe

class _LocData(LocDataBase):

    def __init__(self):
        pass

    def


"""

Region of interest.

This module provides functions for managing regions of interest in localization data.

The Roi class is an object that defines a region of interest within a localization
dataset. It is therefore related to region specifications and a unique LocData object.

The Roi object provides methods for saving all specifications to a yaml file, for loading them, and for returning
LocData with localizations selected to be within the roi region.
"""
import sys
from pathlib import Path
import warnings
from itertools import product
import logging

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import RectangleSelector, PolygonSelector, EllipseSelector
from ruamel.yaml import YAML
from google.protobuf import json_format

from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
import locan.constants
import locan.locan_io.locdata.io_locdata as io
from locan.data.region import *


__all__ = ['Roi', 'rasterize']

logger = logging.getLogger(__name__)


class _MplSelector:  # pragma: no cover
    """
    Class to use matplotlib widgets (RectangleSelector, EllipseSelector or PolygonSelector) on rendered localization
    data.

    Parameters
    ----------
    ax : :class:`matplotlib.axes.Axes`
        Axes to use the widget on.
    type : str
        Type is a string specifying the selector widget that can be either rectangle, ellipse, or polygon.

    Attributes
    ----------
    rois : List of dict
        A list of rois where each element is a dict with keys `region_specs` and 'region'.
        `region_specs` contain a tuple with specifications for the chosen region type (see ``Roi``).
        The region is a string identifier that can be either rectangle, ellipse, or polygon.
    """

    def __init__(self, ax, type='rectangle'):
        self.rois = []
        self.ax = ax

        if type == 'rectangle':
            self.selector = RectangleSelector(self.ax, self.selector_callback, drawtype='box', interactive=True)
            self.type = type

        elif type == 'ellipse':
            self.selector = EllipseSelector(self.ax, self.selector_callback)
            self.type = type

        elif type == 'polygon':
            raise NotImplementedError ('The polygon selection is not working correctly. Use napari.')
        # The PolygonSelector is not working correctly.
        #     self.selector = PolygonSelector(self.ax, self.p_selector_callback)
        #     self.type = type

        else:
            raise TypeError('Type {} is not defined.'.format(type))

        plt.connect('key_press_event', self.key_pressed_callback)

    def selector_callback(self, eclick, erelease):
        """ eclick and erelease are matplotlib events at press and release."""
        print('startposition: {}'.format((eclick.xdata, eclick.ydata)))
        print('endposition  : {}'.format((erelease.xdata, erelease.ydata)))

    def p_selector_callback(self, vertices):
        print('Vertices: {}'.format(vertices))

    def key_pressed_callback(self, event):
        print('Key pressed.')
        if event.key in ['R', 'r']:
            print('RectangleSelector activated.')
            self.selector = RectangleSelector(self.ax, self.selector_callback, drawtype='box', interactive=True)
            self.type = 'rectangle'

        elif event.key in ['E', 'e']:
            print('EllipseSelector activated.')
            self.selector = EllipseSelector(self.ax, self.selector_callback, drawtype='box', interactive=True)
            self.type = 'ellipse'

        elif event.key in ['T', 't']:
            print('PolygonSelector activated.')
            self.selector = PolygonSelector(self.ax, self.p_selector_callback)
            self.type = 'polygon'

        else:
            pass

        if event.key in ['+'] and self.selector.active and self.type == 'rectangle':
            print('Roi was added.')
            region_specs = (np.flip(self.selector.geometry[:, 0]), self.selector.extents[1]-self.selector.extents[0],
                            self.selector.extents[3]-self.selector.extents[2], 0.)
            self.rois.append({'region_specs': region_specs, 'region': self.type})
            print('rois: {}'.format(self.rois))

        elif event.key in ['+'] and self.selector.active and self.type == 'ellipse':
            print('Roi was added.')
            region_specs = (self.selector.center, self.selector.extents[1]-self.selector.extents[0],
                            self.selector.extents[3]-self.selector.extents[2], 0.)

            self.rois.append({'region_specs': region_specs, 'region': self.type})
            print('rois: {}'.format(self.rois))

        elif event.key in ['+'] and self.selector.active and self.type == 'polygon':
            print('Roi was added.')
            vertices_ = self.selector.verts.append(self.selector.verts[0])
            self.rois.append({'region_specs': vertices_, 'region': self.type})
            print('rois: {}'.format(self.rois))

        else:
            pass


class Roi:
    """
    Class for a region of interest on LocData (roi).

    Roi objects define a region of interest for a referenced LocData object.

    Parameters
    ----------
    region : Region
        Geometrical region of interest.
    reference : LocData, dict, locan.data.metadata_pb2.Metadata, None
        Reference to localization data for which the region of interests are defined. It can be a LocData object,
        a reference to a saved SMLM file, or None for indicating no specific reference.
        When referencing a saved SMLM file, reference must be a dict or locan.data.metadata_pb2.Metadata with keys
        `file.path` and `file.type` for a path pointing to a localization file and an integer or string
        indicating the file type.
        Integer or string should be according to locan.constants.FileType.
    loc_properties : tuple of str
        Localization properties in LocData object on which the region selection will be applied (for instance the
        coordinate_labels).


    Attributes
    ----------
    region : Region
        Geometrical region of interest.
    reference : LocData, locan.data.metadata_pb2.Metadata, None
        Reference to localization data for which the region of interests are defined. It can be a LocData object,
        a reference (locan.data.metadata_pb2.Metadata) to a saved SMLM file, or None for indicating no specific reference.
        When referencing a saved SMLM file, reference as attributes `file.path`
        and `file.type` for a path pointing to a localization file and an integer indicating the file type.
        The integer should be according to locan.data.metadata_pb2.Metadata.file.type.
    loc_properties : tuple of str
        Localization properties in LocData object on which the region selection will be applied (for instance the
        coordinate_labels).
    """

    def __init__(self, region, reference=None, loc_properties=()):
        if isinstance(reference, dict):
            self.reference = metadata_pb2.Metadata()
            self.reference.file.path = str(reference['file_path'])
            ft_ = reference['file_type']

            if isinstance(reference['file_type'], int):
                self.reference.file.type = ft_
            elif isinstance(reference['file_type'], str):
                self.reference.file.type = locan.constants.FileType[ft_.upper()].value
            elif isinstance(reference['file_type'], locan.constants.FileType):
                self.reference.file.type = ft_
            elif isinstance(reference['file_type'], metadata_pb2):
                self.reference.file.type = ft_.file.type
            else:
                raise TypeError
        elif isinstance(reference, metadata_pb2.Metadata):
            self.reference = metadata_pb2.Metadata()
            self.reference.MergeFrom(reference)
        else:
            self.reference = reference

        self.region = region
        self.loc_properties = loc_properties

    def __repr__(self):
        return f'Roi(reference={self.reference}, ' \
            f'region={repr(self.region)}, ' \
            f' loc_properties={self.loc_properties})'

    @property
    def region(self):
        return self._region

    @region.setter
    def region(self, region_):
        if not isinstance(region_, Region):
            raise TypeError("An instance of locan.Region must be provided.")
        self._region = region_


    def to_yaml(self, path=None):
        """
        Save Roi object in yaml format.

        Parameters
        ----------
        path : str, bytes, os.PathLike, None
            Path for yaml file. If None a roi file path is generated from the metadata.
        """

        # prepare path
        if path is None and isinstance(self.reference, LocData):
            _file_path = Path(self.reference.meta.file.path)
            _roi_file = _file_path.stem + '_roi.yaml'
            _path = _file_path.with_name(_roi_file)
        elif path is None and isinstance(self.reference, metadata_pb2.Metadata):
            _file_path = Path(self.reference.file.path)
            _roi_file = _file_path.stem + '_roi.yaml'
            _path = _file_path.with_name(_roi_file)
        else:
            _path = Path(path)

        # prepare reference for yaml representation - reference to LocData cannot be represented
        if self.reference is None:
            reference_for_yaml = None
        elif isinstance(self.reference, LocData):
            if self.reference.meta.file.path:
                meta_ = metadata_pb2.Metadata()
                meta_.file.path = self.reference.meta.file.path
                meta_.file.type = self.reference.meta.file.type
                reference_for_yaml = json_format.MessageToJson(meta_, including_default_value_fields=False)
            else:
                warnings.warn('The localization data has to be saved and the file path provided, '
                              'or the reference is lost.', UserWarning)
                reference_for_yaml = None
        else:
            reference_for_yaml = json_format.MessageToJson(self.reference, including_default_value_fields=False)

        region_for_yaml = repr(self.region)
        loc_properties_for_yaml = self.loc_properties

        yaml = YAML()
        output = dict(reference=reference_for_yaml,
                      region=region_for_yaml,
                      loc_properties=loc_properties_for_yaml)
        yaml.dump(output, _path)

    @classmethod
    def from_yaml(cls, path):
        """
        Read Roi object from yaml format.

        Parameters
        ----------
        path : str, bytes, os.PathLike
            Path for yaml file.
        """
        yaml = YAML(typ='safe')
        with open(path) as file:
            yaml_output = yaml.load(file)

        if yaml_output['reference'] is not None:
            reference_ = metadata_pb2.Metadata()
            reference_ = json_format.Parse(yaml_output['reference'], reference_)
        else:
            reference_ = yaml_output['reference']

        region_ = eval(yaml_output['region'])
        loc_properties_ = yaml_output['loc_properties']

        return cls(reference=reference_, region=region_, loc_properties=loc_properties_)

    def locdata(self, reduce=True):
        """
        Localization data according to roi specifications.

        The ROI is applied on locdata properties as specified in self.loc_properties or by taking the first
        applicable locdata.coordinate_labels.

        Parameters
        ----------
        reduce : Bool
            Return the reduced LocData object or keep references alive.

        Returns
        -------
        LocData
            A new instance of LocData with all localizations within region of interest.
        """
        local_parameter = locals()

        if isinstance(self.reference, LocData):
            locdata = self.reference
        elif isinstance(self.reference, metadata_pb2.Metadata):
            locdata = io.load_locdata(self.reference.file.path, self.reference.file.type)
        else:
            raise AttributeError('Valid reference to locdata is missing.')

        if not len(locdata):
            logger.warning('Locdata in reference is empty.')
            new_locdata = locdata

        else:
            if self.loc_properties:
                pfr = self.loc_properties
            else:
                pfr = locdata.coordinate_labels[0:self.region.dimension]

            points = locdata.data[list(pfr)].values
            indices_inside = self._region.contains(points)

            locdata_indices_to_keep = locdata.data.index[indices_inside]
            new_locdata = LocData.from_selection(locdata=locdata, indices=locdata_indices_to_keep)
            if pfr == new_locdata.coordinate_labels:
                new_locdata.region = self._region
            else:
                new_locdata.region = None

        # finish
        if reduce:
            new_locdata.reduce()

        # update metadata
        meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                             parameter=local_parameter,
                             meta=None)
        new_locdata.meta = meta_

        return new_locdata


class RoiLegacy_0:
    """
    Class for a region of interest on LocData (roi).

    Roi objects define a region of interest for a referenced LocData object.

    Parameters
    ----------
    reference : LocData, dict, locan.data.metadata_pb2.Metadata, None
        Reference to localization data for which the region of interests are defined. It can be a LocData object,
        a reference to a saved SMLM file, or None for indicating no specific reference.
        When referencing a saved SMLM file, reference must be a dict or locan.data.metadata_pb2.Metadata with keys `file_path`
        and `file_type` for a path pointing to a localization file and an integer or string indicating the file type.
        Integer or string should be according to locan.constants.FileType.
    region_type : str
        A string indicating the roi shape.
        In 1D it can be `interval`.
        In 2D it can be either `rectangle`, `ellipse`, or closed `polygon`.
        In 3D it can be either `cuboid` or `ellipsoid` or `polyhedron` (not implemented yet).
    region_specs : tuple
        1D rois are defined by the following tuple:
        * interval: (start, stop)
        2D rois are defined by the following tuples:
        * rectangle: ((corner_x, corner_y), width, height, angle)
        * ellipse: ((center_x, center_y), width, height, angle)
        * polygon: ((point1_x, point1_y), (point2_x, point2_y), ..., (point1_x, point1_y))
        3D rois are defined by the following tuples:
        * cuboid: ((corner_x, corner_y, corner_z), length, width, height, angle_1, angle_2, angle_3)
        * ellipsoid: ((center_x, center_y, center_z), length, width, height, angle_1, angle_2, angle_3)
        * polyhedron: (...)
    properties_for_roi : tuple of str
        Localization properties in LocData object on which the region selection will be applied (for instance the
        coordinate_labels).


    Attributes
    ----------
    reference : LocData, locan.data.metadata_pb2.Metadata, None
        Reference to localization data for which the region of interests are defined. It can be a LocData object,
        a reference (locan.data.metadata_pb2.Metadata) to a saved SMLM file, or None for indicating no specific reference.
        When referencing a saved SMLM file, reference as attributes `file_path`
        and `file_type` for a path pointing to a localization file and an integer indicating the file type.
        The integer should be according to locan.data.metadata_pb2.Metadata.file_type.
    _region : RoiRegion, list of RoiRegion
        Object specifying the geometrical region of interest. In case a list of RoiRegion is provided it is the union
        that makes up the region of interest.
    properties_for_roi : tuple of str
        Localization properties in LocData object on which the region selection will be applied (for instance the
        coordinate_labels).

    Warnings
    --------
    `RoiLegacy` is deprecated and should only be used to read legacy _roi.yaml files. Use :class:`locan.Roi` instead.
    """

    def __init__(self, region_type, region_specs, reference=None, properties_for_roi=()):
        if isinstance(reference, dict):
            self.reference = metadata_pb2.Metadata()
            self.reference.file.path = str(reference['file_path'])
            ft_ = reference['file_type']

            if isinstance(reference['file_type'], int):
                self.reference.file.type = ft_
            elif isinstance(reference['file_type'], str):
                self.reference.file.type = locan.constants.FileType[ft_.upper()].value
            elif isinstance(reference['file_type'], locan.constants.FileType):
                self.reference.file.type = ft_
            elif isinstance(reference['file_type'], metadata_pb2):
                self.reference.file.type = ft_.file.type
            else:
                raise TypeError

        elif isinstance(reference, metadata_pb2.Metadata):
            self.reference = metadata_pb2.Metadata()
            self.reference.MergeFrom(reference)
        else:
            self.reference = reference

        self._region = RoiRegion(region_type=region_type, region_specs=region_specs)
        self.properties_for_roi = properties_for_roi

    def __repr__(self):
        return f'Roi(reference={self.reference}, ' \
            f'region={self._region.region_type}, ' \
            f'region_specs={self._region.region_specs},' \
            f' loc_properties={self.properties_for_roi})'

    @property
    def region(self):
        return self._region

    @region.setter
    def region(self, region_):
        if isinstance(region_, RoiRegion):
            self._region = region_

    def to_yaml(self, path=None):
        """
        Save Roi object in yaml format.

        Parameters
        ----------
        path : str, bytes, os.PathLike, None
            Path for yaml file. If None a roi file path is generated from the metadata.
        """
        warnings.warn(
            "RoiLegacy.to_yaml is deprecated, use Roi.to_yaml instead",
            DeprecationWarning
        )

        # prepare path
        if path is None and isinstance(self.reference, LocData):
            _file_path = Path(self.reference.meta.file.path)
            _roi_file = _file_path.stem + '_roi.yaml'
            _path = _file_path.with_name(_roi_file)
        elif path is None and isinstance(self.reference, metadata_pb2.Metadata):
            _file_path = Path(self.reference.file.path)
            _roi_file = _file_path.stem + '_roi.yaml'
            _path = _file_path.with_name(_roi_file)
        else:
            _path = Path(path)

        # prepare reference for yaml representation - reference to LocData cannot be represented
        if self.reference is None:
            reference_for_yaml = None
        elif isinstance(self.reference, LocData):
            if self.reference.meta.file.path:
                meta_ = metadata_pb2.Metadata()
                meta_.file.path = self.reference.meta.file.path
                meta_.file.type = self.reference.meta.file.type
                reference_for_yaml = json_format.MessageToJson(meta_, including_default_value_fields=False)
            else:
                warnings.warn('The localization data has to be saved and the file path provided, '
                              'or the reference is lost.', UserWarning)
                reference_for_yaml = None
        else:
            reference_for_yaml = json_format.MessageToJson(self.reference, including_default_value_fields=False)

        # prepare points for yaml representation - numpy.float has to be converted to float
        def nested_change(iterable, func):
            if isinstance(iterable, (list, tuple, np.ndarray)):
                return [nested_change(x, func) for x in iterable]
            return func(iterable)

        region_specs_for_yaml = nested_change(self._region.region_specs, float)

        region_type_for_yaml = self._region.region_type
        properties_for_roi_for_yaml = self.properties_for_roi

        yaml = YAML()
        output = dict(reference=reference_for_yaml,
                      region_type=region_type_for_yaml,
                      region_specs=region_specs_for_yaml,
                      properties_for_roi=properties_for_roi_for_yaml)
        yaml.dump(output, _path)

    @classmethod
    def from_yaml(cls, path):
        """
        Read Roi object from yaml format.

        Parameters
        ----------
        path : str, bytes, os.PathLike
            Path for yaml file.
        """
        yaml = YAML(typ='safe')
        with open(path) as file:
            yaml_output = yaml.load(file)

        if yaml_output['reference'] is not None:
            reference_ = metadata_pb2.Metadata()
            reference_ = json_format.Parse(yaml_output['reference'], reference_)
        else:
            reference_ = yaml_output['reference']

        region_type_ = yaml_output['region_type']
        region_specs_ = yaml_output['region_specs']
        properties_for_roi_ = yaml_output['properties_for_roi']

        return cls(reference=reference_, region_type=region_type_, region_specs=region_specs_,
                   properties_for_roi=properties_for_roi_)

    def locdata(self, reduce=True):
        """
        Localization data according to roi specifications.

        The ROI is applied on locdata properties as specified in self.loc_properties or by taking the first
        applicable locdata.coordinate_labels.

        Parameters
        ----------
        reduce : Bool
            Return the reduced LocData object or keep references alive.

        Returns
        -------
        LocData
            A new instance of LocData with all localizations within region of interest.
        """
        local_parameter = locals()

        if isinstance(self.reference, LocData):
            locdata = self.reference
        elif isinstance(self.reference, metadata_pb2.Metadata):
            locdata = io.load_locdata(self.reference.file_path, self.reference.file_type)
        else:
            raise AttributeError('Valid reference to locdata is missing.')

        if not len(locdata):
            logger.warning('Locdata in reference is empty.')
            new_locdata = locdata

        else:
            if self.properties_for_roi:
                pfr = self.properties_for_roi
            else:
                pfr = locdata.coordinate_labels[0:self._region.dimension]

            points = locdata.data[list(pfr)].values
            indices_inside = self._region.contains(points)

            locdata_indices_to_keep = locdata.data.index[indices_inside]
            new_locdata = LocData.from_selection(locdata=locdata, indices=locdata_indices_to_keep)
            if pfr == new_locdata.coordinate_labels:
                new_locdata.region = self._region
            else:
                new_locdata.region = None

        # finish
        if reduce:
            new_locdata.reduce()

        # update metadata
        meta_ = _modify_meta(locdata, new_locdata, function_name=sys._getframe().f_code.co_name,
                             parameter=local_parameter,
                             meta=None)
        new_locdata.meta = meta_

        return new_locdata


# todo generalize to take all loc_properties
def rasterize(locdata, support=None, n_regions=(2, 2, 2), loc_properties=()):
    """
    Provide regions of interest by dividing the locdata support in equally sized rectangles.

    Parameters
    ----------
    locdata : LocData
        The localization data from which to select localization data.
    support : tuple of tuples, None
        Coordinate intervals that are divided in `n_regions` subintervals.
        For None intervals are taken from the bounding box.
    n_regions : tuple with size 2 or 3.
        Number of regions in each dimension. E.g. `n_regions` = (2, 2) returns 4 rectangular Roi objects.
    loc_properties : tuple of str
        Localization properties in LocData object on which the region selection will be applied.
        (Only implemented for coordinates labels)

    Returns
    -------
    tuple(locan.data.rois.Roi)
        A sequence of Roi objects
    """
    if len(locdata) == 0:
        raise ValueError('Not implemented for empty LocData objects.')

    if not set(loc_properties).issubset(locdata.coordinate_labels):
        raise ValueError('loc_properties must be tuple with coordinate labels.')

    if loc_properties:
        coordinate_labels_indices = [locdata.coordinate_labels.index(pfr) for pfr in loc_properties]
    else:
        coordinate_labels_indices = range(len(n_regions))

    # specify support
    if support is None:
        support_ = locdata.bounding_box.vertices[coordinate_labels_indices]
        if len(locdata.bounding_box.width) == 0:
            widths = np.zeros(len(n_regions))
        else:
            widths = locdata.bounding_box.width[coordinate_labels_indices] / n_regions
    else:
        support_ = support
        widths = np.diff(support_).flatten() / n_regions

    # specify interval corners
    corners = [np.linspace(*support_d, n_regions_d, endpoint=False)
               for support_d, n_regions_d in zip(support_, n_regions)]
    corners = product(*corners)

    # specify regions
    if len(n_regions) == 2:
        regions = [Rectangle(corner, *widths, 0) for corner in corners]

    elif len(n_regions) == 3:
        raise NotImplementedError('Computation for 3D has not been implemented, yet.')

    else:
        raise ValueError('The shape of n_regions is incompatible.')

    new_rois = tuple([Roi(reference=locdata, region=reg, loc_properties=loc_properties)
                      for reg in regions])
    return new_rois



"""

File input/output for localization data in Thunderstorm files.

"""
from __future__ import annotations
import logging

import pandas as pd

from locan.data.locdata import LocData
import locan.constants
from locan.data import metadata_pb2
from locan.locan_io.locdata.utilities import convert_property_types, open_path_or_file_like, convert_property_names


__all__ = ['load_thunderstorm_header', 'load_thunderstorm_file', 'save_thunderstorm_csv']

logger = logging.getLogger(__name__)


def _read_thunderstorm_header(file):
    """
    Read csv header from a Thunderstorm single-molecule localization file and identify column names.

    Parameters
    ----------
    file : file-like
        Thunderstorm file to read.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the Thunderstorm identifiers.
    """
    # read csv header
    header = file.readline().split('\n')[0]

    # list identifiers
    identifiers = [x.strip('"') for x in header.split(',')]

    column_keys = convert_property_names(properties=identifiers, property_mapping=locan.constants.THUNDERSTORM_KEYS)
    return column_keys


def load_thunderstorm_header(path):
    """
    Load csv header from a Thunderstorm single-molecule localization file and identify column names.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a Thunderstorm file to load.

    Returns
    -------
    list of str
        A list of valid dataset property keys as derived from the Thunderstorm identifiers.
    """
    # read csv header
    with open_path_or_file_like(path) as file:
        return _read_thunderstorm_header(file)


def load_thunderstorm_file(path, nrows=None, convert=True, **kwargs):
    """
    Load data from a Thunderstorm single-molecule localization file.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a Thunderstorm file to load.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with open_path_or_file_like(path) as file:
        columns = _read_thunderstorm_header(file)
        dataframe = pd.read_csv(file, sep=',', skiprows=0, nrows=nrows, names=columns, **kwargs)

    if convert:
        dataframe = convert_property_types(dataframe, types=locan.constants.PROPERTY_KEYS)

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.THUNDERSTORM
    dat.meta.file.path = str(path)

    for property_ in sorted(list(set(columns).intersection({'position_x', 'position_y', 'position_z'}))):
        dat.meta.localization_properties.add(name=property_, unit="nm", type="float")

    del dat.meta.history[:]
    dat.meta.history.add(name='load_thundestorm_file', parameter='path={}, nrows={}'.format(path, nrows))

    return dat


def save_thunderstorm_csv(locdata, path):
    """
    Save LocData attributes Thunderstorm-readable csv-file.

    In the Thunderstorm csv-file file format we store only localization data with Thunderstorm-readable column names.

    Parameters
    ----------
    locdata : LocData
        The LocData object to be saved.
    path : str, bytes, os.PathLike, file-like
        File path including file name to save to.
    """
    # get data from locdata object
    dataframe = locdata.data

    # create reverse mapping to Thunderstorm columns
    inv_map = {v: k for k, v in locan.constants.THUNDERSTORM_KEYS.items()}

    # rename columns
    dataframe = dataframe.rename(index=str, columns=inv_map, inplace=False)

    # write to csv
    dataframe.to_csv(path, float_format='%.10g', index=False)



"""

Compute statistics for localization data.

These values can represent new properties of locdata.

"""
from collections import namedtuple

import numpy as np
import pandas as pd

from locan.data.locdata import LocData


__all__ = ['statistics', 'ranges', 'range_from_collection']


def statistics(locdata, statistic_keys=('count', 'min', 'max', 'mean', 'median', 'std', 'sem')):
    """
    Compute selected statistical parameter for localization data.

    Parameters
    ----------
    locdata : LocData, pandas.DataFrame, pandas.Series
        Localization data

    statistic_keys : str, tuple of str
        Pandas statistic functions. Default: ('count', 'min', 'max', 'mean', 'median', 'std', 'sem')

    Returns
    -------
    dict
        A dict with descriptive statistics.
    """

    if isinstance(locdata, LocData):
        data = locdata.data
    elif isinstance(locdata, (pd.DataFrame, pd.Series)):
        data = locdata
    else:
        raise TypeError('locdata should be of type locan.LocData or pandas.DataFrame.')

    statistics_ = data.agg(statistic_keys)

    if isinstance(locdata, pd.Series):
        p = data.name
        if isinstance(statistic_keys, str):
            dict_ = {p + '_' + statistic_keys: statistics_}
        else:
            dict_ = {p + '_' + s: statistics_[s] for s in statistic_keys}
    else:
        if isinstance(statistic_keys, str):
            generator = (p for p in list(data))
            dict_ = {p + '_' + statistic_keys: statistics_[p] for p in generator}
        else:
            generator = ((p, s) for p in list(data) for s in statistic_keys)
            dict_ = {p + '_' + s: statistics_[p][s] for p, s in generator}

    return dict_


def ranges(locdata: LocData, loc_properties=None, special=None, epsilon=1):
    """
    Provide data ranges for locdata.data property.
    If LocData is empty None is returned.
    If LocData carries a single value, the range will be (value, value + `epsilon`).

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : str, tuple[str], list[str], True, None.
        Localization properties for which the range is determined.
        If None the ranges for all spatial coordinates are returned.
        If True the ranges for all locdata.data properties are returned.
    special : None, str
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from all combined data are returned.
    epsilon : float
        number to specify the range for single values in locdata.

    Returns
    -------
    numpy.ndarray of float with shape (dimension, 2), None
        The data range (min, max) for each localization property.
    """
    if locdata.data.empty:
        return None
    elif len(locdata) == 1:
        pass

    if loc_properties is None:
        ranges_ = locdata.bounding_box.hull.T.copy()
    elif loc_properties is True:
        ranges_ = np.array([locdata.data.min(), locdata.data.max()]).T
    elif isinstance(loc_properties, str):
        ranges_ = np.array([[locdata.data[loc_properties].min(),
                            locdata.data[loc_properties].max()]])
    else:
        loc_properties = list(loc_properties)
        ranges_ = np.array([locdata.data[loc_properties].min(),
                            locdata.data[loc_properties].max()]).T

    if len(locdata) == 1:
        if ranges_.size == 0:
            ranges_ = np.concatenate([locdata.coordinates, locdata.coordinates + epsilon], axis=0).T
        else:
            ranges_ = ranges_ + [0, epsilon]

    if special is None:
        pass
    elif special == 'zero':
        ranges_[:, 0] = 0
    elif special == 'link':
        minmax = np.array([ranges_[:, 0].min(axis=0), ranges_[:, 1].max(axis=0)])
        ranges_ = np.repeat(minmax[None, :], len(ranges_), axis=0)
    else:
        raise ValueError(f'The parameter special={special} is not defined.')

    return ranges_


def range_from_collection(locdatas, loc_properties=None, special=None, epsilon=1):
    """
    Compute the maximum range from all combined localizations for each dimension.

    Parameters
    ----------
    locdatas : list of LocData
        Collection of localization datasets.
    loc_properties : str, tuple[str], list[str], True, None.
        Localization properties for which the range is determined.
        If None the ranges for all spatial coordinates are returned.
        If True the ranges for all locdata.data properties are returned.
    special : None, str
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from all combined data are returned.
    epsilon : float
        number to specify the range for single values in locdata.

    Returns
    -------
    namedtuple
        A namedtuple('Ranges', locdata.coordinate_labels) of namedtuple('Range', 'min max').
    """
    ranges_ = [ranges(locdata=locdata, loc_properties=loc_properties, special=special, epsilon=epsilon)
               for locdata in locdatas]

    mins = np.min([rand[:, 0] for rand in ranges_], axis=0)
    maxs = np.max([rand[:, 1] for rand in ranges_], axis=0)

    if loc_properties is None:
        labels = locdatas[0].coordinate_labels
    else:
        labels = loc_properties

    Ranges = namedtuple('Ranges', labels)
    Range = namedtuple('Range', 'min max')
    result = Ranges(*(Range(min_value, max_value) for min_value, max_value in zip(mins, maxs)))
    return result



"""

This module provides functions for rendering `LocData` objects in 3D.

"""
import logging

import numpy as np
from matplotlib import pyplot as plt
import matplotlib.colors as mcolors

from locan import locdata_id
from locan.data import LocData
from locan.constants import RenderEngine
from locan.configuration import COLORMAP_CONTINUOUS, RENDER_ENGINE
from locan.dependencies import HAS_DEPENDENCY
from locan.data.aggregate import histogram, Bins, _check_loc_properties
from locan.render.transform import adjust_contrast

if HAS_DEPENDENCY["napari"]: import napari


__all__ = ['render_3d', 'render_3d_napari', 'scatter_3d_mpl', 'render_3d_rgb_napari']

logger = logging.getLogger(__name__)


def render_3d_napari(locdata, loc_properties=None, other_property=None,
                     bins=None, n_bins=None, bin_size=10, bin_edges=None, bin_range=None,
                     rescale=None, viewer=None, cmap='viridis', **kwargs):
    """
    Render localization data into a 3D image by binning x,y,z-coordinates into regular bins.
    Render the data using napari.

    Parameters
    ----------
    locdata : LocData
        Localization data.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of `locdata` are used.
    other_property : str, None
        Localization property (columns in `locdata.data`) that is averaged in each pixel.
        If None, localization counts are shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray[float] with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of `locdata` coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple[tuples[float]] with shape (dimension, 2), str, None
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, Trafo, callable, bool, None
        Transformation as defined in :class:`locan.constants.Trafo` or by transformation function.
        For False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer : napari.Viewer
        The viewer object on which to add the image
    cmap : str, Colormap
        The Colormap object used to map normalized data values to RGBA colors.
    kwargs : dict
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer
    """
    if not HAS_DEPENDENCY["napari"]:
        raise ImportError('Function requires napari.')

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning('Locdata carries a single localization.')
        return viewer

    hist = histogram(locdata, loc_properties, other_property,
                     bins, n_bins, bin_size, bin_edges, bin_range,
                     )

    viewer.add_image(hist.data, name=f'LocData {locdata_id}', colormap=cmap, **kwargs)
    return viewer


def render_3d(locdata, render_engine=RENDER_ENGINE, **kwargs):
    """
    Wrapper function to render localization data into a 3D image.
    For complete signatures see render_3d_mpl or corresponding functions.
    """
    if HAS_DEPENDENCY["napari"] and render_engine == RenderEngine.NAPARI:
        return render_3d_napari(locdata, **kwargs)
    else:
        raise NotImplementedError(f"render_3d is not implemented for {render_engine}.")


def scatter_3d_mpl(locdata, ax=None, index=True, text_kwargs=None, **kwargs):
    """
    Scatter plot of locdata elements with text marker for each element.

    Parameters
    ----------
    locdata : LocData
       Localization data.
    ax : matplotlib.axes.Axes3D
       The axes on which to show the plot
    index : bool
       Flag indicating if element indices are shown.
    text_kwargs : dict
       Keyword arguments for :func:`matplotlib.axes.Axes.text`.
    kwargs : dict
       Other parameters passed to :func:`matplotlib.axes.Axes.scatter`.

    Returns
    -------
    matplotlib.axes.Axes
       Axes object with the image.
    """
    if text_kwargs is None:
        text_kwargs = {}

    # Provide matplotlib.axes.Axes if not provided
    if ax is None:
        ax = plt.gca()

    # return ax if no or single point in locdata
    if len(locdata) < 2:
        if len(locdata) == 1:
            logger.warning('Locdata carries a single localization.')
        return ax

    coordinates = locdata.coordinates
    ax.scatter(*coordinates.T, **dict({'marker': '+', 'color': 'grey'}, **kwargs))

    # plot element number
    if index:
        for centroid, marker in zip(coordinates, locdata.data.index.values):
            ax.text(*centroid, marker, **dict({'color': 'grey', 'size': 20}, **text_kwargs))

    ax.set(
           xlabel='position_x',
           ylabel='position_y',
           zlabel='position_z'
           )

    return ax


def render_3d_rgb_napari(locdatas, loc_properties=None, other_property=None,
                      bins=None, n_bins=None, bin_size=10, bin_edges=None, bin_range=None,
                      rescale=None,
                      viewer=None,
                      **kwargs):
    """
    Render localization data into a 3D RGB image by binning x,y,z-coordinates into regular bins.

    Note
    ----
    For rescale=False no normalization is carried out image intensities are clipped to (0, 1) for float value
    or (0, 255) for integer values according to the matplotlib.imshow behavior.
    For rescale=None we apply a normalization to (min, max) of all intensity values.
    For all other rescale options the normalization is applied to each individual image.

    Parameters
    ----------
    locdatas : list[LocData]
        Localization data.
    loc_properties : list, None
        Localization properties to be grouped into bins. If None The coordinate_values of `locdata` are used.
    other_property : str, None
        Localization property (columns in `locdata.data`) that is averaged in each pixel.
        If None localization counts are shown.
    bins : int, sequence, Bins, boost_histogram.axis.Axis, None
        The bin specification as defined in :class:`Bins`
    bin_edges : tuple, list, numpy.ndarray[float] with shape (dimension, n_bin_edges), None
        Array of bin edges for all or each dimension.
    n_bins : int, list, tuple, numpy.ndarray, None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float, list, tuple, numpy.ndarray, None
        The size of bins in units of locdata coordinate units for all or each dimension.
        5 would describe bin_size of 5 for all bins in all dimensions.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other dimension.
        To specify arbitrary sequence of `bin_sizes` use `bin_edges` instead.
    bin_range : tuple, tuple of tuples of float with shape (dimension, 2), str, None
        The data bin_range to be taken into consideration for all or each dimension.
        ((min_x, max_x), (min_y, max_y), ...) bin_range for each coordinate;
        for None (min, max) bin_range are determined from data;
        for 'zero' (0, max) bin_range with max determined from data.
    rescale : int, str, Trafo, callable, bool, None
        Transformation as defined in :class:`locan.constants.Trafo` or by transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min intensities
        For 'equal' intensity values are rescaled by histogram equalization.
    viewer : napari.Viewer
        The viewer object on which to add the image
    kwargs : dict
        Other parameters passed to :func:`napari.Viewer.add_image`.

    Returns
    -------
    napari.Viewer
    """
    if not HAS_DEPENDENCY["napari"]:
        raise ImportError('Function requires napari.')

    # Provide napari viewer if not provided
    if viewer is None:
        viewer = napari.Viewer()

    locdata_temp = LocData.concat(locdatas)

    # return viewer if no or single point in locdata
    if len(locdata_temp) < 2:
        if len(locdata_temp) == 1:
            logger.warning('Locdata carries a single localization.')
        return viewer

    if bin_edges is None:
        _, bins, labels = histogram(locdata_temp, loc_properties, other_property,
                                       bins, n_bins, bin_size, bin_edges, bin_range)
    else:
        labels = _check_loc_properties(locdata_temp, loc_properties)
        bins = Bins(bin_edges=bin_edges, labels=labels)

    imgs = [histogram(locdata, loc_properties, other_property, bin_edges=bins.bin_edges
                      ).data
            for locdata in locdatas
            ]

    if rescale is None:
        norm = mcolors.Normalize(vmin=np.min(imgs), vmax=np.max(imgs))
    else:
        norm = rescale
    imgs = [adjust_contrast(img, rescale=norm) for img in imgs]

    new = np.zeros_like(imgs[0])
    rgb_stack = np.stack([new] * 3, axis=3)
    for i, img in enumerate(imgs):
        rgb_stack[:, :, :, i] = img
    rgb_stack = np.transpose(rgb_stack, axes=(2, 1, 0, 3))

    viewer.add_image(rgb_stack, name=f'LocData {locdata_id}', rgb=True, **kwargs)
    return viewer



from pathlib import Path

import matplotlib.colors as mcolors

from locan.configuration import DATASETS_DIR, RENDER_ENGINE, N_JOBS, \
    COLORMAP_CONTINUOUS, COLORMAP_DIVERGING, COLORMAP_CATEGORICAL
from locan.constants import RenderEngine


def test_datasets_dir():
    # print(DATASETS_DIR)
    assert isinstance(DATASETS_DIR, Path)


def test_render_enginge():
    assert isinstance(RENDER_ENGINE, RenderEngine)


def test_n_jobs():
    assert isinstance(N_JOBS, int)


def test_colormaps():
    colormaps = [COLORMAP_CONTINUOUS, COLORMAP_DIVERGING, COLORMAP_CATEGORICAL]
    for item in colormaps:
        assert isinstance(item, mcolors.Colormap) or isinstance(mcolors.Colormap(item), mcolors.Colormap)



import pytest
import numpy as np
import matplotlib.pyplot as plt  # this import is needed for interactive tests

from locan import RenderEngine  # this import is needed for interactive tests
from locan.dependencies import HAS_DEPENDENCY
from locan import render_2d_mpl, render_2d_scatter_density, scatter_2d_mpl, render_2d_rgb_mpl
from locan import render_2d, apply_window
from locan import cluster_dbscan, transform_affine


def test_render_2d_mpl_empty(locdata_empty):
    render_2d_mpl(locdata_empty, bin_size=5)
    # plt.show()

    plt.close('all')


def test_render_2d_mpl_single(locdata_single_localization, caplog):
    render_2d_mpl(locdata_single_localization, bin_size=0.5)
    assert caplog.record_tuples == [('locan.render.render2d', 30, 'Locdata carries a single localization.')]
    # plt.show()

    plt.close('all')


def test_render_2d_mpl(locdata_blobs_2d):
    render_2d_mpl(locdata_blobs_2d)
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=[[500, 1000], [500, 1000]], cbar=False)

    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale=None)
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale=True)
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale='unity')
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale='equal')
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range=None, rescale=(0, 50))
    render_2d_mpl(locdata_blobs_2d, bin_size=100, bin_range='zero')

    fig, ax = plt.subplots(nrows=1, ncols=2)
    render_2d_mpl(locdata_blobs_2d, ax=ax[0])
    render_2d_mpl(locdata_blobs_2d, bin_range='zero', ax=ax[1])

    render_2d_mpl(locdata_blobs_2d, ax=ax[0], colorbar_kws=dict(ax=ax[0]))
    render_2d_mpl(locdata_blobs_2d, bin_range='zero', ax=ax[1])

    plt.close('all')


@pytest.mark.visual
# this is to check overlay of rendered image and single localization points
def test_render_2d_mpl_show(locdata_blobs_2d):
    # print(locdata_blobs_2d.coordinates)
    render_2d_mpl(locdata_blobs_2d, bin_size=10, bin_range=None, rescale=None, other_property='position_y')
    # plt.plot(locdata_blobs_2d.coordinates[:, 0], locdata_blobs_2d.coordinates[:, 1], 'o')
    plt.show()

    plt.close('all')


@pytest.mark.skipif(not HAS_DEPENDENCY["mpl_scatter_density"], reason="requires mpl_scatter_density")
def test_render_2d_scatter_density(locdata_blobs_2d):
    render_2d_scatter_density(locdata_blobs_2d)
    # render_2d_scatter_density(locdata_blobs_2d, bin_range=[[500, 1000], [500, 1000]], cbar=False)

    # render_2d_scatter_density(locdata_blobs_2d, bin_range=None, vmin=0, vmax=1)

    # fig, ax = plt.subplots(nrows=1, ncols=2)
    # render_2d_scatter_density(locdata_blobs_2d, ax=ax[0])
    # render_2d_scatter_density(locdata_blobs_2d, bin_range='zero', ax=ax[1])
    #
    # render_2d_scatter_density(locdata_blobs_2d, ax=ax[0], colorbar_kws=dict(ax=ax[0]))
    # render_2d_scatter_density(locdata_blobs_2d, bin_range='zero', ax=ax[1])

    # render_2d_scatter_density(locdata_blobs_2d, other_property='position_x')

    # plt.show()

    plt.close('all')


@pytest.mark.skipif(not HAS_DEPENDENCY["mpl_scatter_density"], reason="requires mpl_scatter_density")
def test_render_2d_scatter_density_empty(locdata_empty):
    render_2d_scatter_density(locdata_empty)
    # plt.show()

    plt.close('all')


@pytest.mark.skipif(not HAS_DEPENDENCY["mpl_scatter_density"], reason="requires mpl_scatter_density")
def test_render_2d_scatter_density_single(locdata_single_localization, caplog):
    render_2d_scatter_density(locdata_single_localization)
    assert caplog.record_tuples == [('locan.render.render2d', 30, 'Locdata carries a single localization.')]
    # plt.show()
    plt.close('all')


@pytest.mark.gui
@pytest.mark.parametrize("test_input, expected", list((member, 0) for member in list(RenderEngine)))
def test_render_2d(locdata_blobs_2d, test_input, expected):
    if HAS_DEPENDENCY["napari"] and test_input == RenderEngine.NAPARI:
        render_2d(locdata_blobs_2d, render_engine=test_input)
        # napari.run()
    else:
        render_2d(locdata_blobs_2d, render_engine=test_input)
    # plt.show()

    plt.close('all')


def test_scatter_2d_mpl(locdata_2d):
    scatter_2d_mpl(locdata_2d, text_kwargs=dict(color='r'), color='r')
    # plt.show()

    plt.close('all')

def test_scatter_2d_mpl_empty(locdata_empty):
    scatter_2d_mpl(locdata_empty)
    # plt.show()

    plt.close('all')


def test_scatter_2d_mpl_single(locdata_single_localization, caplog):
    scatter_2d_mpl(locdata_single_localization)
    assert caplog.record_tuples == [('locan.render.render2d', 30, 'Locdata carries a single localization.')]
    # plt.show()

    plt.close('all')


@pytest.mark.visual  # Visual check repeating previously checked functionality
def test_scatter_2d_mpl_2(locdata_blobs_2d):
    _, collection = cluster_dbscan(locdata_blobs_2d, eps=20, min_samples=3)
    render_2d_mpl(locdata_blobs_2d)
    scatter_2d_mpl(collection)
    plt.show()

    plt.close('all')


def test_apply_window():
    img = np.ones((10, 10))
    img_filtered = apply_window(image=img, window_function='tukey', alpha=0.4)
    assert np.array_equal(img_filtered[0, :], np.zeros(10))
    assert np.array_equal(img_filtered[:, 0], np.zeros(10))


def test_render_2d_rgb_mpl_empty(locdata_empty):
    render_2d_rgb_mpl([locdata_empty, locdata_empty], bin_size=1)
    # plt.show()

    plt.close('all')


def test_render_2d_rgb_mpl_single(locdata_empty, locdata_single_localization, caplog):
    render_2d_rgb_mpl([locdata_empty, locdata_single_localization], bin_size=1)
    assert caplog.record_tuples == [('locan.render.render2d', 30, 'Locdata carries a single localization.')]
    # plt.show()

    plt.close('all')


def test_render_2d_rgb_mpl(locdata_2d):
    render_2d_rgb_mpl([locdata_2d, locdata_2d], bin_size=1)
    # plt.show()

    plt.close('all')


@pytest.mark.visual  # Visual check repeating previously checked functionality
def test_render_2d_rgb_mpl_2(locdata_blobs_2d):
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=20)
    plt.show()

    plt.close('all')


@pytest.mark.visual  # Visual check repeating previously checked functionality
def test_render_2d_rgb_mpl_3(locdata_blobs_2d):
    """Check intensity normalization."""
    locdata_0 = locdata_blobs_2d
    locdata_1 = transform_affine(locdata_blobs_2d, offset=(20, 0))

    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=20)
    plt.show()
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=20, rescale=False)
    plt.show()

    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=100)
    plt.show()
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=100, rescale=True)
    plt.show()
    render_2d_rgb_mpl([locdata_0, locdata_1], bin_size=100, rescale=False)
    plt.show()

    plt.close('all')



from pathlib import Path
import tempfile
import pickle

import pytest
from scipy import stats

from locan.analysis.analysis_base import _Analysis, _list_parameters
from locan.analysis import metadata_analysis_pb2


def test_Analysis():
    ae = _Analysis(limits=(100, 110), meta={'comment': 'this is an example'})
    # print(repr(ae))
    assert str(ae) == repr(ae)
    assert ae.parameter == {'limits': (100, 110)}

    #print(ae.meta)
    #print(dir(metadata_analysis_pb2))
    assert isinstance(ae.meta, metadata_analysis_pb2.AMetadata)
    assert ae.meta.identifier
    assert ae.meta.creation_time
    assert ae.meta.method.name == "_Analysis"
    assert ae.meta.comment == 'this is an example'

    ae.meta.comment = 'new comment'
    assert ae.meta.comment == 'new comment'

    ae.meta.map['variable key'] = 'value_1'
    assert str(ae.meta.map) == "{'variable key': 'value_1'}"

    ae.meta.map['key_2'] = 'value_2'
    assert ae.meta.map['key_2'] == 'value_2'

    with pytest.raises(NotImplementedError):
        ae.compute()


def test_pickling_Analysis():
    ae = _Analysis(limits=(100, 110), meta={'comment': 'this is an example'})
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / 'pickled_analysis.pickle'
        with open(file_path, 'wb') as file:
            pickle.dump(ae, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, 'rb') as file:
            ae_pickled = pickle.load(file)
        assert ae_pickled.parameter == ae.parameter
        assert isinstance(ae_pickled.meta, metadata_analysis_pb2.AMetadata)
        assert ae_pickled.meta == ae.meta


def test__list_parameters():
    with pytest.raises(AttributeError):
        parameters = _list_parameters('fail')
    parameters = _list_parameters('expon')
    assert parameters == ['loc', 'scale']
    parameters = _list_parameters(stats.gamma)
    assert parameters == ['a', 'loc', 'scale']
    parameters = _list_parameters(stats.poisson)
    assert parameters == ['mu', 'loc']



# from locan.utils.miscellaneous import iterate_2d_array
from locan import iterate_2d_array


def test_iterate_2d_array():
    assert list(iterate_2d_array(1, 1)) == [(0, 0)]
    assert list(iterate_2d_array(1, 3)) == [(0, 0)]
    assert list(iterate_2d_array(2, 2)) == [(0, 0), (0, 1)]
    assert list(iterate_2d_array(3, 2)) == [(0, 0), (0, 1), (1, 0)]
    assert list(iterate_2d_array(2, 3)) == [(0, 0), (0, 1)]
    assert list(iterate_2d_array(3, 1)) == [(0, 0), (1, 0), (2, 0)]
    assert list(iterate_2d_array(0, 1)) == []
